from typing import Callable, Tuple

import holoviews
from holoviews import Store
from holoviews.streams import Buffer, Pipe
import judo
import numpy
import pandas
import pandas as pd
from scipy.interpolate import griddata

from fragile.core.api_classes import Callback


class StreamingPlot:
    """Represents a holoviews plot updated with streamed data."""

    name = ""
    default_opts = {"framewise": True, "axiswise": True, "normalize": True, "shared_axes": False}
    stream_class = Pipe
    default_bokeh_opts = {}
    default_mpl_opts = {}

    def __init__(
        self,
        plot: Callable,
        stream=None,
        data=None,
        bokeh_opts: dict = None,
        mpl_opts: dict = None,
        **kwargs,
    ):
        """
        Initialize a :class:`StreamingPlot`.

        Args:
            plot: Callable that returns a holoviews plot.
            stream: Class used to stream data to the plot.
            data: Passed to :class:`Plot`.``get_plot_data``. Contains the necessary data to
                initialize the plot.
            args: Passed to ``opts``.
            kwargs: Passed to ``opts``.
        """
        self.data_stream = None
        self.plot = None
        bokeh_opts = bokeh_opts if bokeh_opts is not None else {}
        mpl_opts = mpl_opts if mpl_opts is not None else {}
        self.bokeh_opts = {**self.default_bokeh_opts, **bokeh_opts}
        self.mpl_opts = {**self.default_mpl_opts, **mpl_opts}
        self.common_kwargs = {**self.default_opts, **kwargs}
        self.init_stream(stream, data)
        self.init_plot(plot)

    @property
    def opts_kwargs(self):
        if Store.current_backend == "bokeh":
            backend_kwargs = self.bokeh_opts
        elif Store.current_backend == "matplotlib":
            backend_kwargs = self.mpl_opts
        else:
            backend_kwargs = {}
        return {**self.common_kwargs, **backend_kwargs}

    @opts_kwargs.setter
    def opts_kwargs(self, kwargs):
        self.common_kwargs = {**self.common_kwargs, **kwargs}

    def get_default_data(self):
        raise NotImplementedError()

    def get_default_stream(self, data):
        return self.stream_class(data=data)

    def preprocess_data(self, data):
        """Perform the necessary data wrangling for plotting the data."""
        return data

    def send(self, data) -> None:
        """Stream data to the plot and keep track of how many times the data has been streamed."""
        data = self.preprocess_data(data)
        self.data_stream.send(data)

    def init_plot(self, plot: Callable) -> None:
        """
        Initialize the holoviews plot to accept streaming data.

        Args:
            plot: Callable that returns a holoviews plot.

        """

        def plot_func(data):
            _plot = plot(data)
            return self.opts(plot=_plot)

        self.plot = holoviews.DynamicMap(plot_func, streams=[self.data_stream])
        self.opts()

    def init_stream(self, stream=None, data=None):
        """Initialize the data stream that will be used to stream data to the plot."""
        if stream is None:
            data = self.preprocess_data(data) if data is not None else self.get_default_data()
            stream = self.get_default_stream(data=data)
        self.data_stream = stream

    def update_kwargs(self, **kwargs):
        """Update the supplied options kwargs with backend specific parameters."""
        if Store.current_backend == "bokeh":
            opt_dict = dict(self.bokeh_opts)
        elif Store.current_backend == "matplotlib":
            opt_dict = dict(self.mpl_opts)
        else:
            opt_dict = {}
        opt_dict.update(kwargs)
        return opt_dict

    def opts(self, plot=None, **kwargs):
        """Update the plot parameters. Same as `holoviews` `opts`."""
        if self.plot is None:
            return
        self.common_kwargs.update(kwargs)
        if plot is None:
            self.plot = self.plot.opts(**self.opts_kwargs)
        else:
            return plot.opts(**self.opts_kwargs)


class Div(StreamingPlot):
    default_opts = {
        "framewise": True,
        "axiswise": True,
        "normalize": True,
    }

    name = "Div"
    default_bokeh_opts = {
        "height": 350,
        "width": 350,
    }

    def __init__(
        self,
        data=None,
        plot=holoviews.Div,
        **kwargs,
    ):
        """
        Initialize a :class:`Table`.

        Args:
            data: Data to initialize the stream.
            stream: :class:`holoviews.stream` type. Defaults to :class:`Pipe`.
            bokeh_opts: Default options for the plot when rendered using the "bokeh" backend.
            mpl_opts: Default options for the plot when rendered using the "matplotlib" backend.
            **kwargs: Passed to :class:`StreamingPlot`.

        """
        super(Div, self).__init__(
            plot=plot,
            data=data,
            **kwargs,
        )

    def get_default_data(self):
        return ""

    def opts(self, plot=None, **kwargs):
        """Update the plot parameters. Same as `holoviews` `opts`."""
        if self.plot is None:
            return
        self.common_kwargs.update(kwargs)
        if plot is None:
            self.plot = self.plot.opts(holoviews.opts.Div(**self.opts_kwargs))
        else:
            return plot.opts(holoviews.opts.Div(**self.opts_kwargs))


class Table(StreamingPlot):
    """``holoviews.Table`` with data streaming capabilities."""

    default_opts = {
        "framewise": True,
        "axiswise": True,
        "normalize": True,
    }

    name = "table"
    default_bokeh_opts = {
        "height": 350,
        "width": 350,
    }

    def __init__(
        self,
        data=None,
        plot=holoviews.Table,
        **kwargs,
    ):
        """
        Initialize a :class:`Table`.

        Args:
            data: Data to initialize the stream.
            stream: :class:`holoviews.stream` type. Defaults to :class:`Pipe`.
            bokeh_opts: Default options for the plot when rendered using the "bokeh" backend.
            mpl_opts: Default options for the plot when rendered using the "matplotlib" backend.
            **kwargs: Passed to :class:`StreamingPlot`.

        """
        super(Table, self).__init__(
            plot=plot,
            data=data,
            **kwargs,
        )

    def get_default_data(self):
        return pandas.DataFrame()

    def opts(self, plot=None, **kwargs):
        """Update the plot parameters. Same as `holoviews` `opts`."""
        if self.plot is None:
            return
        self.common_kwargs.update(kwargs)
        if plot is None:
            self.plot = self.plot.opts(holoviews.opts.Table(**self.opts_kwargs))
        else:
            return plot.opts(holoviews.opts.Table(**self.opts_kwargs))


class RGB(StreamingPlot):
    """``holoviews.RGB`` with data streaming capabilities."""

    name = "rgb"
    default_bokeh_opts = {"xaxis": None, "yaxis": None}

    def __init__(self, data=None, plot=holoviews.RGB, **kwargs):
        """Initialize a :class:`RGB`."""
        super(RGB, self).__init__(plot=plot, data=data, **kwargs)

    def get_default_data(self):
        return []

    def opts(self, plot=None, **kwargs):
        """Update the plot parameters. Same as `holoviews` `opts`."""
        if self.plot is None:
            return
        self.common_kwargs.update(kwargs)
        if plot is None:
            self.plot = self.plot.opts(holoviews.opts.RGB(**self.opts_kwargs))
        else:
            return plot.opts(holoviews.opts.RGB(**self.opts_kwargs))


class Curve(StreamingPlot):
    """
    Create a ``holoviews.Curve`` plot that plots steaming data.

    The streaming process is handled using a :class:`Buffer`.
    """

    name = "curve"
    default_bokeh_opts = {
        "height": 350,
        "width": 400,
        "shared_axes": False,
        "tools": ["hover"],
    }
    stream_class = Buffer

    def __init__(
        self,
        data=None,
        plot=holoviews.Curve,
        buffer_length: int = 10000,
        index: bool = False,
        data_names=("x", "y"),
        **kwargs,
    ):
        """
        Initialize a :class:`Curve`.

        Args:
            buffer_length: Maximum number of data points that will be displayed in the plot.
            index: Passed to the :class:`Buffer` that streams data to the plot.
            data: Passed to :class:`Plot`.``get_plot_data``. Contains the necessary data to \
                  initialize the plot.
            bokeh_opts: Default options for the plot when rendered using the \
                       "bokeh" backend.
            mpl_opts: Default options for the plot when rendered using the \
                    "matplotlib" backend.
        """

        self._buffer_length = buffer_length
        self._index = index

        self._data_names = data.columns.values if data is not None else data_names
        super(Curve, self).__init__(
            plot=plot,
            data=data,
            **kwargs,
        )

    def get_default_stream(self, data):
        return self.stream_class(data=data, length=self._buffer_length, index=self._index)

    def get_default_data(self):
        return pandas.DataFrame(columns=self._data_names)

    def opts(self, plot=None, **kwargs):
        """Update the plot parameters. Same as `holoviews` `opts`."""
        if self.plot is None:
            return
        self.common_kwargs.update(kwargs)
        if plot is None:
            self.plot = self.plot.opts(holoviews.opts.Curve(**self.opts_kwargs))
        else:
            return plot.opts(holoviews.opts.Curve(**self.opts_kwargs))


class Histogram(StreamingPlot):
    """
    Create a ``holoviews.Histogram`` plot that plots steaming data.

    The streaming process is handled using a :class:`Pipe`.
    """

    name = "histogram"
    default_opts = {
        "ylabel": "count",
        "framewise": True,
        "axiswise": True,
        "normalize": True,
        "shared_axes": False,
    }
    default_bokeh_opts = {"tools": ["hover"]}

    def __init__(
        self,
        data=None,
        plot=None,
        n_bins: int = 20,
        **kwargs,
    ):
        """
        Initialize a :class:`Histogram`.

        Args:
            n_bins: Number of bins of the histogram that will be plotted.
            data: Used to initialize the plot.
            bokeh_opts: Default options for the plot when rendered using the \
                       "bokeh" backend.
            mpl_opts: Default options for the plot when rendered using the \
                    "matplotlib" backend.
        """
        self.n_bins = n_bins
        self.xlim = (None, None)
        super(Histogram, self).__init__(
            plot=self.plot_histogram if plot is None else plot,
            data=data,
            **kwargs,
        )

    @staticmethod
    def plot_histogram(data):
        """
        Plot the histogram.

        Args:
            data: Tuple containing (values, bins), xlim. xlim is a tuple \
                  containing two typing_.Scalars that represent the limits of the x \
                  axis of the histogram.

        Returns:
            Histogram plot.

        """
        plot_data, xlim = data
        return holoviews.Histogram(plot_data).redim(x=holoviews.Dimension("x", range=xlim))

    def opts(self, plot=None, **kwargs):
        """Update the plot parameters. Same as `holoviews` `opts`."""
        if self.plot is None:
            return
        self.common_kwargs.update(kwargs)
        if plot is None:
            self.plot = self.plot.opts(holoviews.opts.Histogram(**self.opts_kwargs))
        else:
            return plot.opts(holoviews.opts.Histogram(**self.opts_kwargs))

    def preprocess_data(self, data):
        data = judo.to_numpy(data)
        if data is None or not data.shape:
            data = numpy.zeros(10)
            data[-1] = 1
        if numpy.isinf(data).all() or numpy.isnan(data).all():
            data = numpy.zeros_like(data)
            data[-1] = 1
        else:
            not_inf_data = data[numpy.logical_not(numpy.isinf(data))]
            data[numpy.logical_and(numpy.isinf(data), data > 0)] = not_inf_data.max() + 10
            data[numpy.logical_and(numpy.isinf(data), data < 0)] = not_inf_data.min() - 10
            data[numpy.isnan(data)] = 0.0
        return numpy.histogram(data, self.n_bins), self.xlim

    def get_default_data(self):
        data = numpy.arange(10)
        return self.preprocess_data(data)


class Bivariate(StreamingPlot):
    """
    Create a ``holoviews.Bivariate`` plot that plots steaming data.

    The streaming process is handled using a :class:`Pipe`.
    """

    name = "bivariate"
    default_bokeh_opts = {
        "height": 350,
        "width": 400,
        "tools": ["hover"],
    }

    def __init__(self, data=None, plot=holoviews.Bivariate, **kwargs):
        """
        Initialize a :class:`Bivariate`.

        Args:
            data: Passed to ``holoviews.Bivariate``.
            bokeh_opts: Default options for the plot when rendered using the \
                       "bokeh" backend.
            mpl_opts: Default options for the plot when rendered using the \
                    "matplotlib" backend.
            *args: Passed to ``holoviews.Bivariate``.
            **kwargs: Passed to ``holoviews.Bivariate``.
        """
        super(Bivariate, self).__init__(
            plot=plot,
            data=data,
            **kwargs,
        )

    def opts(self, plot=None, **kwargs):
        """Update the plot parameters. Same as `holoviews` `opts`."""
        if self.plot is None:
            return
        self.common_kwargs.update(kwargs)
        if plot is None:
            self.plot = self.plot.opts(holoviews.opts.Bivariate(**self.opts_kwargs))
        else:
            return plot.opts(holoviews.opts.Bivariate(**self.opts_kwargs))

    def __opts(
        self,
        title="",
        xlabel: str = "x",
        ylabel: str = "y",
        framewise: bool = True,
        axiswise: bool = True,
        normalize: bool = True,
        *args,
        **kwargs,
    ):
        """
        Update the plot parameters. Same as ``holoviews`` ``opts``.

        The default values updates the plot axes independently when being \
        displayed in a :class:`Holomap`.
        """
        kwargs = self.update_kwargs(**kwargs)
        # Add specific defaults to Scatter
        scatter_kwargs = dict(kwargs)
        if Store.current_backend == "bokeh":
            scatter_kwargs["size"] = scatter_kwargs.get("size", 3.5)
        elif Store.current_backend == "matplotlib":
            scatter_kwargs["s"] = scatter_kwargs.get("s", 15)
        self.plot = self.plot.opts(
            holoviews.opts.Bivariate(
                title=title,
                xlabel=xlabel,
                ylabel=ylabel,
                framewise=framewise,
                axiswise=axiswise,
                normalize=normalize,
                *args,
                **kwargs,
            ),
            holoviews.opts.Scatter(
                alpha=0.7,
                xlabel=xlabel,
                ylabel=ylabel,
                framewise=framewise,
                axiswise=axiswise,
                normalize=normalize,
                *args,
                **scatter_kwargs,
            ),
            holoviews.opts.NdOverlay(
                normalize=normalize,
                framewise=framewise,
                axiswise=axiswise,
            ),
        )

    def get_default_data(self):
        return []


class QuadMesh(StreamingPlot):
    """
    Create a ``holoviews.Histogram`` plot that plots steaming data.

    The streaming process is handled using a :class:`Pipe`.
    """

    name = "quadmesh"
    default_bokeh_opts = {
        "tools": ["hover"],
        "bgcolor": "lightgray",
        "colorbar": True,
        "height": 350,
        "line_width": 1.0,
        "width": 400,
        "shared_axes": False,
        "cmap": "viridis",
    }

    # default_mpl_opts = {"linewidth": 1.0}

    def __init__(
        self,
        data=None,
        plot=None,
        n_points: int = 20,
        **kwargs,
    ):
        """
        Initialize a :class:`Histogram`.

        Args:
            n_bins: Number of bins of the histogram that will be plotted.
            data: Used to initialize the plot.
        """
        self.n_points = n_points
        super(QuadMesh, self).__init__(
            plot=self.plot_quadmesh if plot is None else plot,
            data=data,
            **kwargs,
        )

    def send(self, data, xx=None, yy=None, zz=None) -> None:
        """Stream data to the plot and keep track of how many times the data has been streamed."""
        data = self.preprocess_data(data, xx=xx, yy=yy, zz=zz)
        self.data_stream.send(data)

    def plot_quadmesh(self, data):
        """
        Plot the data as an energy landscape.

        Args:
            data: (x, y, xx, yy, z, xlim, ylim). x, y, z represent the \
                  coordinates of the points that will be interpolated. xx, yy \
                  represent the meshgrid used to interpolate the points. xlim, \
                  ylim are tuples containing the limits of the x and y axes.

        Returns:
            Plot representing the interpolated energy landscape of the target points.

        """
        xx, yy, zz, xlim, ylim, zlim = data
        plot = holoviews.QuadMesh((xx, yy, zz))
        return plot
        return plot.redim(
            x=holoviews.Dimension("x", range=xlim),
            y=holoviews.Dimension("y", range=ylim),
            z=holoviews.Dimension("z", range=zlim),
        )

    def preprocess_data(
        self,
        data: Tuple[numpy.ndarray, numpy.ndarray, numpy.ndarray],
        xx=None,
        yy=None,
        zz=None,
    ):
        """Create the meshgrid needed to interpolate the target data points."""
        x, y, z = (data[:, 0], data[:, 1], data[:, 2]) if isinstance(data, numpy.ndarray) else data
        x, y, z = judo.to_numpy(x), judo.to_numpy(y), judo.to_numpy(z)
        # target grid to interpolate to
        if zz is None:
            xi = numpy.linspace(x.min(), x.max(), self.n_points)
            yi = numpy.linspace(y.min(), y.max(), self.n_points)
            xx, yy = numpy.meshgrid(xi, yi)
            zz = griddata((x, y), z, (xx, yy), method="linear")
        xlim, ylim, zlim = (x.min(), x.max()), (y.min(), y.max()), (z.min(), z.max())
        return xx, yy, zz, xlim, ylim, zlim

    def get_default_data(self):
        X = numpy.random.standard_normal((10, 2))
        z = numpy.random.standard_normal(10)
        data = X[:, 0], X[:, 1], z
        return self.preprocess_data(data)

    def opts(self, plot=None, **kwargs):
        """Update the plot parameters. Same as `holoviews` `opts`."""
        if self.plot is None:
            return
        self.common_kwargs.update(kwargs)
        if plot is None:
            self.plot = self.plot.opts(holoviews.opts.QuadMesh(**self.opts_kwargs))
        else:
            return plot.opts(holoviews.opts.QuadMesh(**self.opts_kwargs))


class QuadMeshContours(QuadMesh):
    default_opts = {"shared_axes": True}
    default_bokeh_opts = {
        "tools": ["hover"],
        "bgcolor": "lightgray",
        "alpha": 0.9,
        "show_legend": False,
        "height": 350,
        "width": 400,
        "line_color": "black",
        # "shared_axes": False,
        "colorbar": True,
        "cmap": "viridis",  # ["black"],
    }

    def __init__(self, data=None, levels: int = 16, **kwargs):
        self.levels = levels
        super(QuadMeshContours, self).__init__(data=data, **kwargs)

    def plot_quadmesh(self, data):
        """
        Plot the data as an energy landscape.

        Args:
            data: (x, y, xx, yy, z, xlim, ylim). x, y, z represent the \
                  coordinates of the points that will be interpolated. xx, yy \
                  represent the meshgrid used to interpolate the points. xlim, \
                  ylim are tuples containing the limits of the x and y axes.

        Returns:
            Plot representing the interpolated energy landscape of the target points.

        """
        xx, yy, zz, xlim, ylim, zlim = data
        mesh = holoviews.QuadMesh((xx, yy, zz))
        plot = holoviews.operation.contours(mesh, levels=self.levels)
        return (mesh * plot).redim(
            x=holoviews.Dimension("x", range=xlim),
            y=holoviews.Dimension("y", range=ylim),
            z=holoviews.Dimension("z", range=zlim),
        )

    def send(self, data, xx=None, yy=None, zz=None) -> None:
        x, y, z = data
        self.common_kwargs["zlim"] = (z.min(), z.max())
        super(QuadMeshContours, self).send(data=data, xx=xx, yy=yy, zz=zz)

    def opts(self, plot=None, **kwargs):
        """Update the plot parameters. Same as `holoviews` `opts`."""
        if self.plot is None:
            return
        self.common_kwargs.update(kwargs)
        if plot is None:
            self.plot = self.plot.opts(
                holoviews.opts.Contours(**{**self.opts_kwargs, **{"cmap": ["black"]}}),
                holoviews.opts.QuadMesh(**self.opts_kwargs),
            )
        else:
            return plot.opts(
                holoviews.opts.Contours(**{**self.opts_kwargs, **{"cmap": ["black"]}}),
                holoviews.opts.QuadMesh(**self.opts_kwargs),
            )


class Scatter(StreamingPlot):
    default_bokeh_opts = {
        "tools": ["hover"],
        "bgcolor": "lightgray",
        "line_color": "black",
        # "fill_color": "red",
        "height": 350,
        "size": 3.5,
        "width": 400,
        "shared_axes": False,
    }
    default_mpl_opts = {"s": 1}

    def __init__(
        self,
        data=None,
        plot=holoviews.Scatter,
        n_points: int = 20,
        **kwargs,
    ):
        """
        Initialize a :class:`Histogram`.

        Args:
            n_bins: Number of bins of the histogram that will be plotted.
            data: Used to initialize the plot.
        """
        self.n_points = n_points
        super(Scatter, self).__init__(
            plot=plot,
            data=data,
            **kwargs,
        )

    def get_default_data(self):
        return pd.DataFrame({"x": [], "y": []})

    def opts(self, plot=None, **kwargs):
        """Update the plot parameters. Same as `holoviews` `opts`."""
        if self.plot is None:
            return
        self.common_kwargs.update(kwargs)
        if plot is None:
            self.plot = self.plot.opts(holoviews.opts.Scatter(**self.opts_kwargs))
        else:
            return plot.opts(holoviews.opts.Scatter(**self.opts_kwargs))


class Landscape2D(StreamingPlot):
    """
    Plots the interpolated landscaped of values of a set of points.

    The data is visualized creating a :class:`holoviews.QuadMesh` with a \
    :class:`holoviews.Contours` plot with the original data points displayed as \
    a :class:`holoviews.Scatter`.
    """

    name = "landscape"
    default_bokeh_opts = {
        "height": 350,
        "width": 400,
        "tools": ["hover"],
        "shared_axes": False,
    }

    def __init__(
        self,
        data=None,
        contours: bool = True,
        **kwargs,
    ):
        """
        Initialize a :class:`Landscape2d`.

        Args:
            n_points: Number of points per dimension used to create the \
                      mesh-grid grid that will be used to interpolate the data.
            data: Initial data for the plot.
            bokeh_opts: Default options for the plot when rendered using the \
                       "bokeh" backend.
            mpl_opts: Default options for the plot when rendered using the \
                    "matplotlib" backend.
            invert_cmap: If ``True``, invert the colormap to assign high value \
                         colors to the lowest values.

        """
        self.quadmesh_ps = QuadMesh(data=data, **kwargs)
        self.contours_ps = (
            None
            if not contours
            else QuadMeshContours(data=data, stream=self.quadmesh_ps.data_stream, **kwargs)
        )
        self.scatter_ps = Scatter(data=data)
        super(Landscape2D, self).__init__(
            stream=self.quadmesh_ps.data_stream,
            plot=self.plot_landscape,
            data=data,
            **kwargs,
        )

    def init_plot(self, plot: Callable) -> None:
        """
        Initialize the holoviews plot to accept streaming data.

        Args:
            plot: Callable that returns a holoviews plot.

        """
        # self.plot = self.quadmesh_ps.plot
        if True:  # self.contours_ps is not None:
            self.plot = self.contours_ps.plot

        # self.plot = self.plot * self.scatter_ps.plot

    def plot_landscape(self, data):
        """
        Plot the data as an energy landscape.

        Args:
            data: (x, y, xx, yy, z, xlim, ylim). x, y, z represent the \
                  coordinates of the points that will be interpolated. xx, yy \
                  represent the meshgrid used to interpolate the points. xlim, \
                  ylim are tuples containing the limits of the x and y axes.

        Returns:
            Plot representing the interpolated energy landscape of the target points.

        """
        plot = self.quadmesh_ps.plot_quadmesh(data)
        if self.contours_ps:
            plot = plot * self.contours_ps.plot_quadmesh(data)
        return plot.opts(
            holoviews.opts.QuadMesh(**self.quadmesh_ps.opts_kwargs),
            holoviews.opts.Contours(**self.contours_ps.opts_kwargs),
        )

    def preprocess_data(self, data):
        return self.quadmesh_ps.preprocess_data(data)

    def send(self, data) -> None:
        """Stream data to the plot and keep track of how many times the data has been streamed."""
        self.quadmesh_ps.send(data)
        df = pandas.DataFrame({"x": data[0], "y": data[1]})
        self.scatter_ps.send(df)
        # self.plot = self.plot.opts(framewise=True)
        xlim, ylim = (data[0].min(), data[0].max()), (data[1].min(), data[1].max())
        zlim = (data[2].min(), data[2].max())
        self.quadmesh_ps.plot = self.quadmesh_ps.plot.redim(
            x=holoviews.Dimension("x", range=xlim),
            y=holoviews.Dimension("y", range=ylim),
            z=holoviews.Dimension("z", range=zlim),
        )
        self.contours_ps.plot = self.contours_ps.plot.redim(
            x=holoviews.Dimension("x", range=xlim),
            y=holoviews.Dimension("y", range=ylim),
            z=holoviews.Dimension("z", range=zlim),
        )
        # self.plot = self.plot

    def get_default_data(self):
        return self.quadmesh_ps.get_default_data()

    def opts(self, **kwargs):
        self.quadmesh_ps.opts(**kwargs)
        if self.contours_ps is not None:
            self.contours_ps.opts(**kwargs)

    def __opts(
        self,
        title="Distribution landscape",
        xlabel: str = "x",
        ylabel: str = "y",
        framewise: bool = True,
        axiswise: bool = True,
        normalize: bool = True,
        cmap: str = "default",
        **kwargs,
    ):
        """
        Update the plot parameters. Same as ``holoviews`` ``opts``.

        The default values updates the plot axes independently when being \
        displayed in a :class:`Holomap`.
        """
        kwargs = self.update_kwargs(**kwargs)
        cmap = cmap if cmap != "default" else ("viridis_r" if self.invert_cmap else "viridis")
        # Add specific defaults to Contours
        contours_kwargs = dict(kwargs)
        if Store.current_backend == "bokeh":
            contours_kwargs["line_width"] = contours_kwargs.get("line_width", 1)
        elif Store.current_backend == "matplotlib":
            contours_kwargs["linewidth"] = contours_kwargs.get("linewidth", 1)

        # Add specific defaults to Scatter
        scatter_kwargs = dict(kwargs)
        if Store.current_backend == "bokeh":
            scatter_kwargs["fill_color"] = scatter_kwargs.get("fill_color", "red")
            scatter_kwargs["size"] = scatter_kwargs.get("size", 3.5)
        elif Store.current_backend == "matplotlib":
            scatter_kwargs["color"] = scatter_kwargs.get("color", "red")
            scatter_kwargs["s"] = scatter_kwargs.get("s", 15)

        self.plot = self.plot.opts(
            holoviews.opts.QuadMesh(
                cmap=cmap,
                colorbar=True,
                title=title,
                bgcolor="lightgray",
                xlabel=xlabel,
                ylabel=ylabel,
                framewise=framewise,
                axiswise=axiswise,
                normalize=normalize,
                **kwargs,
            ),
            holoviews.opts.Contours(
                cmap=["black"],
                alpha=0.9,
                title=title,
                xlabel=xlabel,
                ylabel=ylabel,
                show_legend=False,
                framewise=framewise,
                axiswise=axiswise,
                normalize=normalize,
                **contours_kwargs,
            ),
            holoviews.opts.Scatter(
                alpha=0.7,
                xlabel=xlabel,
                ylabel=ylabel,
                framewise=framewise,
                axiswise=axiswise,
                normalize=normalize,
                **scatter_kwargs,
            ),
            holoviews.opts.NdOverlay(
                normalize=normalize,
                framewise=framewise,
                axiswise=axiswise,
            ),
        )


class PlotCallback(Callback):
    def __init__(self, report_interval: int = 1, **kwargs):
        self.report_interval = report_interval
        super(PlotCallback, self).__init__(**kwargs)

    def after_evolve(self):
        super(PlotCallback, self).before_walkers()
        if self.swarm.epoch % self.report_interval == 0 or self.swarm.epoch == 0:
            self.send()

    def send(self):
        raise NotImplementedError()

    def panel(self):
        raise NotImplementedError()

    def run_end(self):
        self.send()



"""This module contains the :class:`PlangymEnv` class."""
import copy
from typing import Callable, Optional, Tuple, Union

import judo
from judo import Bounds, dtype, random_state, typing
import numpy
from plangym.core import PlangymEnv as _PlangymEnv

from fragile.core.api_classes import EnvironmentAPI, SwarmAPI
from fragile.core.typing import InputDict, StateData, StateDict, Tensor


class PlangymEnv(EnvironmentAPI):
    r"""
    Wrapper class for running a Swarm to solve planning tasks in a Plangym environment.

    This class allows running gymnasium simulations compatible with the plangym API.
    It can be used as an interface for passing states and actions, and receiving
    observation and reward information from the environment.

    Attributes:
        plangym_env (Env): The Plangym environment this instance is wrapped around.
        _has_terminals (bool): `True` if the environment has terminal signals, `False` otherwise.
        _has_rgb (bool): `True` if the environment includes RGB color data, `False` otherwise.

    Examples:
        >>> import plangym
        >>> plangym_env = plangym.make("CartPole-v0")
        >>> env = PlangymEnv(plangym_env)
    """

    def __init__(self, plangym_env: _PlangymEnv, swarm: Optional[SwarmAPI] = None):
        r"""
        Initialize a new instance of the PlangymEnv class.

        Args:
            plangym_env (_PlangymEnv): The underlying Plangym environment.
            swarm (Optional[SwarmAPI], optional): The swarm to use in this environment.
                Defaults to None.

        Examples:
            >>> import plangym
            >>> plangym_env = plangym.make("CartPole-v0")
            >>> env = PlangymEnv(plangym_env)
        """
        self._plangym_env = plangym_env
        state, obs = plangym_env.reset()
        *_, infos = plangym_env.step(plangym_env.sample_action())
        self._has_rgb = "rgb" in infos
        self.rgb_shape = infos["rgb"].shape if self.has_rgb else None
        states_shape = None if not plangym_env.STATE_IS_ARRAY else state.shape
        states_dtype = type(state) if not plangym_env.STATE_IS_ARRAY else state.dtype
        self._has_terminals = (
            hasattr(self.plangym_env, "possible_to_win") and self.plangym_env.possible_to_win
        )

        self._states_shape = states_shape
        self._states_dtype = states_dtype
        super(PlangymEnv, self).__init__(
            swarm=swarm,
            observs_dtype=obs.dtype,
            observs_shape=obs.shape,
            action_shape=plangym_env.action_space.shape,
            action_dtype=plangym_env.action_space.dtype,
        )

    @property
    def states_shape(self) -> tuple:
        """
        Returns the shape of the states tensor.

        Returns:
            tuple: The shape of the states tensor.

        Examples:
            >>> import plangym
            >>> plangym_env = plangym.make("CartPole-v0")
            >>> env = PlangymEnv(plangym_env)
            >>> env.states_shape
            (4,)
        """
        return self._states_shape

    @property
    def states_dtype(self):
        """
        Returns the data type of the states tensor.

        Returns:
            judo.dtype: The data type of the states tensor.

        Examples:
            >>> import plangym
            >>> plangym_env = plangym.make("CartPole-v0")
            >>> env = PlangymEnv(plangym_env)
            >>> env.states_dtype
            dtype('float64')
        """
        return self._states_dtype

    @property
    def plangym_env(self) -> _PlangymEnv:
        """
        Returns the underlying Plangym environment.

        Returns:
            plangym.PlangymEnv: The underlying :class:`plangym.Plangym` environment.
        """
        return self._plangym_env

    @property
    def inputs(self) -> InputDict:
        """
        Returns a dictionary of input data for the environment.

        Returns:
            InputDict: A dictionary of input data for the environment.

        Examples:
            >>> import plangym
            >>> plangym_env = plangym.make("CartPole-v0")
            >>> env = PlangymEnv(plangym_env)
            >>> env.inputs  # doctest: +NORMALIZE_WHITESPACE
            {'actions': {},
             'states': {'clone': True},
             'dt': {'optional': True, 'default': 1}}
        """
        plangym_inputs = {"states": {"clone": True}, "dt": {"optional": True, "default": 1}}
        return {**super(PlangymEnv, self).inputs, **plangym_inputs}

    @property
    def outputs(self) -> Tuple[str, ...]:
        """
        Returns a tuple of output variables for the environment.

        Returns:
            Tuple[str, ...]: A tuple of output variables for the environment.
        Examples:
            >>> import plangym
            >>> plangym_env = plangym.make("CartPole-v0")
            >>> env = PlangymEnv(plangym_env)
            >>> tuple(sorted(env.outputs))
            ('infos', 'n_steps', 'observs', 'oobs', 'rewards', 'states')
        """
        outputs = ("n_steps", "infos", "states")
        if self._has_terminals:
            outputs = outputs + ("terminals",)
        if self.has_rgb:
            outputs = outputs + ("rgb",)
        return super(PlangymEnv, self).outputs + outputs

    @property
    def param_dict(self) -> StateDict:
        """
        Returns a dictionary of parameters for the environment.

        Returns:
            StateDict: A dictionary of parameters for the environment.

        Examples:
            >>> import plangym
            >>> plangym_env = plangym.make("CartPole-v0")
            >>> env = PlangymEnv(plangym_env)
            >>> print(env.param_dict)  # doctest: +NORMALIZE_WHITESPACE
            {'observs': {'shape': (4,), 'dtype': dtype('float32')},
             'rewards': {'dtype': <class 'numpy.float32'>},
             'oobs': {'dtype': <class 'numpy.bool_'>},
             'actions': {'shape': (), 'dtype': dtype('int64')},
             'n_steps': {'dtype': <class 'numpy.int32'>},
             'infos': {'shape': None, 'dtype': <class 'dict'>},
             'states': {'shape': (4,), 'dtype': dtype('float64')}}
        """
        plangym_params = {
            "n_steps": {"dtype": dtype.int32},
            "infos": {"shape": None, "dtype": dict},
            "states": {"shape": self._states_shape, "dtype": self._states_dtype},
        }
        if self._has_terminals:
            plangym_params["terminals"] = {"dtype": dtype.bool}
        if self.has_rgb:
            plangym_params["rgb"] = {"shape": self.rgb_shape, "dtype": dtype.uint8}
        return {**super(PlangymEnv, self).param_dict, **plangym_params}

    @property
    def has_rgb(self) -> bool:
        """
        Return whether the environment includes RGB color data or not.

        Returns:
            bool: `True` if the environment includes RGB color data, `False` otherwise.
        """
        return self._has_rgb

    def __getattr__(self, item):
        return getattr(self.plangym_env, item)

    def step(self, actions: Tensor, states: Tensor, dt: int = 1) -> StateDict:
        """
        Takes an action in the environment and returns information about the new state.

        Args:
            actions (Tensor): The actions to take. Shape should be (n_walkers,) + actions_shape.
            states (Tensor): The states to act on. Shape should be (n_walkers,) + state_shape.
            dt (int, optional): The number of simulation steps to take per gym step. Defaults to 1.

        Returns:
            Dict[str, Any]: A dictionary containing the following keys:
                - observs (Tensor): The observatons from the last step.
                    Has shape (n_walkers,) + observation_shape.
                - rewards (Tensor): The rewards received from the last step.
                    Has shape (n_walkers,).
                - oobs (List[Any]): List of episode endings from the last step.
                    Length is n_walkers.
                - infos (List[Dict[str, Any]]): Additional information about the last step.
                    Length is n_walkers.
                - n_steps (List[int]): The number of simulation steps taken in the last step.
                    Length is n_walkers.
                - states (Tensor): The new states after taking the given actions.
                    Has shape (n_walkers,) + state_shape.
                - terminals (List[bool], optional): List of raw terminal values if available.
                    Length is n_walkers. Only returned if the environment has terminal signals.
                - rgb (Tensor, optional): The rendered RGB output of the last step. Only returned
                    by some environments. Has shape (n_walkers, h, w, c).

        Examples:
            >>> import plangym
            >>> plangym_env = plangym.make("CartPole-v0")
            >>> env = PlangymEnv(plangym_env)
            >>> n_walkers = 3
            >>> actions = np.array([plangym_env.sample_action() for i in range(n_walkers)])
            >>> state_dict = env.reset(n_walkers=n_walkers, inplace=False)
            >>> result = env.step(actions, state_dict["states"])
            >>> type(result)
            <class 'dict'>
            >>> tuple(sorted(result.keys()))
            ('infos', 'n_steps', 'observs', 'oobs', 'rewards', 'states')
        """
        step_data = self.plangym_env.step_batch(actions=actions, states=states, dt=dt)
        new_states, observs, rewards, oobs, infos = step_data
        n_steps = [info.get("n_steps", 1) for info in infos]
        states_data = dict(
            observs=observs,
            rewards=rewards,
            oobs=oobs,
            infos=infos,
            n_steps=n_steps,
            states=new_states,
        )
        if self._has_terminals:
            terminals = [info["terminal"] for info in infos] if "terminal" in infos[0] else oobs
            states_data["terminals"] = terminals
        if self.has_rgb:
            rgbs = [info["rgb"] for info in infos]
            states_data["rgb"] = rgbs
        return states_data

    def reset(
        self,
        inplace: bool = True,
        root_walker: Optional[StateData] = None,
        states: Optional[StateData] = None,
        n_walkers: Optional[int] = None,
        **kwargs,
    ):
        """
        Resets the environment(s) to its initial state.

        This method resets the states and observables of the environment(s)
        stored in the object's `swarm.state` attribute to their initial values.

        Args:
            inplace (bool): If True, updates the current instance state with
                the reset values. If False, returns a new dictionary with
                the states, observables and info dicts.
            root_walker (Optional[StateData]): The state information to reset from,
                if not using default initial state.
                Defaults to None.
            states (Optional[StateData]): The states to use as initial values, if
                provided, it will ignore `root_walker`. Defaults to None.
            n_walkers (Optional[int]): The number of walkers to reset. Defaults to None.
            **kwargs: Other parameters that might be necessary
                depending on the specific implementation of the class.

        Returns:
            Optional[StateDict]: A StateDict containing the states,
            observables, and info dict after the reset. Only returned
            when `inplace` is False.

        Examples:
            >>> import plangym
            >>> plangym_env = plangym.make("CartPole-v0")
            >>> n_walkers = 3
            >>> # Reset environment and update the state of the swarm:
            >>> env = PlangymEnv(plangym_env, swarm=None)
            >>> env.reset(n_walkers=n_walkers)  # Fails because there is no swarm
            Traceback (most recent call last):
            ...
            AttributeError: 'NoneType' object has no attribute 'state'

            >>> # Get reset data without modifying current instance:
            >>> env = PlangymEnv(plangym_env)
            >>> reset_data = env.reset(n_walkers=n_walkers, inplace=False)
            >>> tuple(sorted(reset_data.keys()))
            ('infos', 'observs', 'rewards', 'states')
        """
        n_walkers = len(self.swarm) if n_walkers is None else n_walkers
        # TODO: add support for defaults in param_dict to avoid this mess.
        default_value = -numpy.inf
        if self.swarm is not None:
            default_value = numpy.inf if self.swarm.minimize else -numpy.inf
        if root_walker is None:
            state, observs = self.plangym_env.reset()
            new_states = [copy.deepcopy(state) for _ in range(n_walkers)]
            new_observs = [copy.deepcopy(observs) for _ in range(n_walkers)]
        else:  # TODO: test this
            new_states = self.get("states")
            new_observs = self.get("observs")
        infos = [{} for _ in range(n_walkers)]
        rewards = judo.ones(n_walkers) * default_value
        if inplace:
            self.update(
                states=new_states,
                observs=new_observs,
                infos=infos,
                rewards=rewards,
                inactives=True,
            )
        else:
            return dict(states=new_states, observs=new_observs, rewards=rewards, infos=infos)


class Function(EnvironmentAPI):
    """
    Environment that represents an arbitrary mathematical function bounded in a \
    given interval.
    """

    default_inputs = {"actions": {}, "observs": {"clone": True}}

    def __init__(
        self,
        function: Callable[[typing.Tensor], typing.Tensor],
        bounds: Union[Bounds, "gym.spaces.box.Box"],
        custom_domain_check: Callable[[typing.Tensor, typing.Tensor, int], typing.Tensor] = None,
        actions_as_perturbations: bool = True,
        start_same_pos: bool = False,
        x0: Tensor = None,
    ):
        """
        Initialize a :class:`Function`.

        Args:
            function: Callable that takes a batch of vectors (batched across \
                      the first dimension of the array) and returns a vector of \
                      typing.Scalar. This function is applied to a batch of walker \
                      observations.
            bounds: :class:`Bounds` that defines the domain of the function.
            custom_domain_check: Callable that checks points inside the bounds \
                    to know if they are in a custom domain when it is not just \
                    a set of rectangular bounds. It takes a batch of points as \
                    input and returns an array of booleans. Each ``True`` value \
                    indicates that the corresponding point is **outside**  the \
                    ``custom_domain_check``.
            actions_as_perturbations: If ``True`` the actions are interpreted as \
                    perturbations that will be applied to the past states. \
                    If ``False`` the actions are interpreted as the new states to \
                    be evaluated.

        """
        if not isinstance(bounds, Bounds) and bounds.__class__.__name__ != "Box":
            raise TypeError(f"Bounds needs to be an instance of Bounds or Box, found {bounds}")
        self.function = function
        self.bounds = bounds if isinstance(bounds, Bounds) else Bounds.from_space(bounds)
        self._action_space = self.bounds.to_space()
        self.custom_domain_check = custom_domain_check
        self._actions_as_perturbations = actions_as_perturbations
        self._x0 = x0
        self.start_same_pos = start_same_pos
        super(Function, self).__init__(
            observs_shape=self.shape,
            observs_dtype=dtype.float32,
            action_dtype=dtype.float32,
            action_shape=self.shape,
        )

    @property
    def n_dims(self) -> int:
        """Return the number of dimensions of the function to be optimized."""
        return len(self.bounds)

    @property
    def shape(self) -> Tuple[int, ...]:
        """Return the shape of the environment."""
        return self.bounds.shape

    @property
    def action_space(self) -> "gym.spaces.box.Box":
        """Action space with the same characteristics as self.bounds."""
        return self._action_space

    @classmethod
    def from_bounds_params(
        cls,
        function: Callable,
        shape: tuple = None,
        high: Union[int, float, typing.Tensor] = numpy.inf,
        low: Union[int, float, typing.Tensor] = numpy.NINF,
        custom_domain_check: Callable[[typing.Tensor], typing.Tensor] = None,
    ) -> "Function":
        """
        Initialize a function defining its shape and bounds without using a :class:`Bounds`.

        Args:
            function: Callable that takes a batch of vectors (batched across \
                      the first dimension of the array) and returns a vector of \
                      typing.Scalar. This function is applied to a batch of walker \
                      observations.
            shape: Input shape of the solution vector without taking into account \
                    the batch dimension. For example, a two-dimensional function \
                    applied to a batch of 5 walkers will have shape=(2,), even though
                    the observations will have shape (5, 2)
            high: Upper bound of the function domain. If it's a typing.Scalar it will \
                  be the same for all dimensions. If it's a numpy array it will \
                  be the upper bound for each dimension.
            low: Lower bound of the function domain. If it's a typing.Scalar it will \
                  be the same for all dimensions. If it's a numpy array it will \
                  be the lower bound for each dimension.
            custom_domain_check: Callable that checks points inside the bounds \
                    to know if they are in a custom domain when it is not just \
                    a set of rectangular bounds.

        Returns:
            :class:`Function` with its :class:`Bounds` created from the provided arguments.

        """
        if (
            not (judo.is_tensor(high) or isinstance(high, (list, tuple)))
            and not (judo.is_tensor(low) or isinstance(low, (list, tuple)))
            and shape is None
        ):
            raise TypeError("Need to specify shape or high or low must be an array.")
        bounds = Bounds(high=high, low=low, shape=shape)
        return Function(function=function, bounds=bounds, custom_domain_check=custom_domain_check)

    def __repr__(self):
        text = "{} with function {}, obs shape {},".format(
            self.__class__.__name__,
            self.function.__name__,
            self.shape,
        )
        return text

    def step(self, actions, observs, **kwargs) -> StateData:
        """

        Sum the target action to the observations to obtain the new points, and \
        evaluate the reward and boundary conditions.

        Returns:
            Dictionary containing the information of the new points evaluated.

             ``{"states": new_points, "observs": new_points, "rewards": typing.Scalar array, \
             "oobs": boolean array}``

        """
        new_points = actions + observs if self._actions_as_perturbations else actions
        rewards = self.function(new_points).flatten()
        oobs = self.calculate_oobs(points=new_points, rewards=rewards)
        return dict(observs=new_points, rewards=rewards, oobs=oobs)

    def reset(
        self,
        inplace: bool = True,
        root_walker: Optional[StateData] = None,
        states: Optional[StateData] = None,
        **kwargs,
    ) -> Union[None, StateData]:
        """
        Reset the :class:`Function` to the start of a new episode and returns \
        an :class:`StatesEnv` instance describing its internal state.
        """
        if root_walker is None:
            new_points = self.sample_action(batch_size=self.swarm.n_actives)
            if self._x0 is not None:
                new_points[:] = self._x0
            elif self.start_same_pos:
                new_points[:] = new_points[0]
            actions = judo.zeros_like(new_points) if self._actions_as_perturbations else new_points
            rewards = self.function(new_points).flatten()
        else:
            new_points = self.get("observs")
            rewards = self.get("rewards")
            actions = (
                self.get("actions") if self._actions_as_perturbations else self.get("observs")
            )
        if inplace:
            self.update(observs=new_points, rewards=rewards, actions=actions)
        else:
            return dict(observs=new_points, rewards=rewards, actions=actions)

    def calculate_oobs(self, points: typing.Tensor, rewards: typing.Tensor) -> typing.Tensor:
        """
        Determine if a given batch of vectors lie inside the function domain.

        Args:
            points: Array of batched vectors that will be checked to lie inside \
                    the :class:`Function` bounds.
            rewards: Array containing the rewards of the current walkers.

        Returns:
            Array of booleans of length batch_size (points.shape[0]) that will \
            be ``True`` if a given point of the batch lies outside the bounds, \
            and ``False`` otherwise.

        """
        oobs = judo.logical_not(self.bounds.contains(points)).flatten()
        if self.custom_domain_check is not None:
            points_in_bounds = judo.logical_not(oobs)
            oobs[points_in_bounds] = self.custom_domain_check(
                points[points_in_bounds],
                rewards[points_in_bounds],
                len(rewards),
            )
        return oobs

    def sample_action(self, batch_size: int) -> typing.Tensor:
        """
        Return a matrix of points sampled uniformly from the :class:`Function` \
        domain.

        Args:
            batch_size: Number of points that will be sampled.

        Returns:
            Array containing ``batch_size`` points that lie inside the \
            :class:`Function` domain, stacked across the first dimension.

        """
        shape = tuple([batch_size])
        if judo.Backend.is_numpy():
            shape = shape + self.shape
        new_points = random_state.uniform(
            low=judo.astype(self.bounds.low, judo.float),
            high=judo.astype(self.bounds.high, judo.float32),
            size=shape,
        )
        return new_points



from typing import Iterable, Optional, Tuple, Union

import einops
import judo
import numpy
import numpy as np

from fragile.callbacks.data_tracking import StoreInitAction
from fragile.core.api_classes import Callback, EnvironmentAPI, PolicyAPI, WalkersAPI
from fragile.core.swarm import Swarm
from fragile.core.typing import InputDict, StateData, StateDict, Tensor


class BoundaryInitAction(Callback):
    name = "boundary_action"
    default_outputs = ("actions",)

    def __init__(self, only_first_epoch: bool = True, **kwargs):
        self.only_first_epoch = only_first_epoch
        super(BoundaryInitAction, self).__init__(**kwargs)

    def after_policy(self):
        assert hasattr(self.swarm.policy, "bounds")
        if self.only_first_epoch and self.swarm.epoch > 0:
            return
        low_val = einops.repeat(self.swarm.policy.bounds.low, "n -> b n", b=self.swarm.n_walkers)
        high_val = einops.repeat(self.swarm.policy.bounds.high, "n -> b n", b=self.swarm.n_walkers)
        condition = judo.random_state.random(low_val.shape) < 0.5
        actions = numpy.where(condition, low_val, high_val)
        self.update(actions=actions)


class FMCPolicy(PolicyAPI):

    default_inputs = {"init_actions": {}, "oobs": {}}

    def __init__(self, inner_swarm, follow_best: bool = False, **kwargs):
        self.inner_swarm = inner_swarm
        self._follow_best_flag = follow_best
        super(FMCPolicy, self).__init__(**kwargs)

    @property
    def param_dict(self) -> StateDict:
        pdict = self.inner_swarm.env.param_dict
        return {**{"init_actions": dict(pdict["actions"])}, **pdict}

    def select_actions(self, **kwargs) -> Union[Tensor, StateData]:
        if self._follow_best_flag:
            return self._follow_best()
        if hasattr(self.inner_swarm.env.action_space, "n"):
            return self._choose_majority()
        init_actions = self.inner_swarm.get("init_actions")
        selected_action = init_actions.mean(0)[np.newaxis, :]
        return selected_action

    def _choose_majority(self):
        init_actions = judo.to_numpy(self.inner_swarm.get("init_actions"))
        y = numpy.bincount(init_actions, minlength=self.inner_swarm.env.action_space.n)
        most_used_action = judo.tensor([y.argmax()])
        return most_used_action

    def _follow_best(self):
        return self.inner_swarm.root.data["init_actions"]
        # init_actions = judo.to_numpy(self.inner_swarm.get("init_actions"))
        # y = judo.to_numpy(self.inner_swarm.get("scores"))
        # oobs, terminals = self.inner_swarm.get("oobs"), self.inner_swarm.get("terminals")
        # index = judo.arange(len(y))
        # valid_ix = judo.logical_or(~oobs, terminals)
        # best_ix = y[valid_ix].argmin() if self.inner_swarm.minimize else y[valid_ix].argmax()
        # best_ix_orig = index[valid_ix][best_ix]
        # best_action = init_actions[best_ix_orig][np.newaxis, :]
        # return best_action


class EnvSwarm(EnvironmentAPI):
    def __init__(self, swarm):
        self.inner_swarm = swarm
        super(EnvSwarm, self).__init__(
            swarm=None,
            action_shape=swarm.env.action_shape,
            action_dtype=swarm.env.action_dtype,
            observs_shape=swarm.env.observs_shape,
            observs_dtype=swarm.env.observs_dtype,
        )

    def __getattr__(self, item):
        return getattr(self.inner_swarm.env, item)

    @property
    def inputs(self) -> InputDict:
        return self.inner_swarm.env.inputs

    @property
    def outputs(self) -> Tuple[str, ...]:
        return self.inner_swarm.env.outputs + ("scores",)

    @property
    def param_dict(self) -> StateDict:
        pdict = self.inner_swarm.env.param_dict
        return {**{"scores": pdict["rewards"]}, **pdict}

    def make_transitions(self, inplace: bool = True, **kwargs) -> Union[None, StateData]:
        env_data = self.swarm.inputs_from_swarm("env")
        env_data = self.inner_swarm.env.make_transitions(inplace=False, **env_data)
        env_data["scores"] = env_data["rewards"]
        if self.inner_swarm.walkers.accumulate_reward:
            env_data["scores"] += self.get("scores")
        if inplace:
            self.swarm.state.update(**env_data)
        else:
            return env_data


class WalkersSwarm(WalkersAPI):
    def __init__(self, inner_swarm, **kwargs):
        self.inner_swarm = inner_swarm
        self.accumulate_reward = self.inner_swarm.walkers.accumulate_reward
        super(WalkersSwarm, self).__init__(**kwargs)

    @property
    def param_dict(self) -> StateDict:
        return self.inner_swarm.param_dict

    def __getattr__(self, item):
        return getattr(self.inner_swarm, item)

    def run_epoch(self, inplace: bool = True, **kwargs) -> StateData:
        self.inner_swarm.run(root_walker=dict(self.swarm.state))
        return {}

    def reset(self, inplace: bool = True, **kwargs):
        super(WalkersSwarm, self).reset(inplace=inplace, **kwargs)
        self.inner_swarm.reset()
        walker = self.inner_swarm.state.export_walker(0)
        self.swarm.state.update(**walker)


class FMCSwarm(Swarm):
    walkers_last = False

    def __init__(
        self,
        swarm: Swarm,
        policy: PolicyAPI = None,
        env: EnvironmentAPI = None,
        callbacks: Optional[Iterable[Callback]] = None,
        minimize: bool = False,
        max_epochs: int = 1e100,
        store_init_action: Callback = None,
    ):
        store_init_action = StoreInitAction() if store_init_action is None else store_init_action
        swarm.register_callback(store_init_action)
        # if hasattr(swarm.policy, "bounds"):
        #    swarm.register_callback(BoundaryInitAction())
        self._swarm = swarm
        policy = FMCPolicy(swarm) if policy is None else policy
        env = EnvSwarm(swarm) if env is None else env
        super(FMCSwarm, self).__init__(
            n_walkers=swarm.n_walkers,
            policy=policy,
            env=env,
            callbacks=callbacks,
            minimize=minimize,
            max_epochs=int(max_epochs),
            walkers=WalkersSwarm(swarm),
        )

    @property
    def swarm(self) -> Swarm:
        return self._swarm

    @property
    def n_walkers(self) -> int:
        return 1

    def inputs_from_swarm(self, name):
        def get_names():
            names = []
            inputs = getattr(self.swarm, name).inputs
            for k, v in inputs.items():
                is_optional = v.get("optional", False)
                if not is_optional or (is_optional and k in self.swarm.param_dict.keys()):
                    names.append(k)
            return names

        return {name: self.state[name] for name in get_names()}



"""Module for optimizing functions using a :class:`Swarm`."""
# from fragile.optimize.models import ESModel
from fragile.core.swarm import FunctionMapper
from fragile.optimize.env import Function, MinimizerWrapper



import panel

from fragile.core.fractalai import relativize
from fragile.dataviz.stream_plots import Histogram, PlotCallback


class PlotWalkers(PlotCallback):
    name = "plot_walkers"

    def __init__(
        self,
        plot_scores: bool = True,
        plot_diversities: bool = True,
        plot_virtual_rewards: bool = True,
        plot_norm_scores: bool = True,
        plot_norm_diversities: bool = True,
        plot_rewards: bool = True,
        **kwargs,
    ):
        super(PlotWalkers, self).__init__(**kwargs)
        self._plot_rewards = plot_rewards
        self._plot_scores = plot_scores
        self._plot_diversities = plot_diversities
        self._plot_virtual_rewards = plot_virtual_rewards
        self._plot_norm_scores = plot_norm_scores
        self._plot_norm_diversities = plot_norm_diversities
        self.virtual_reward_sp = None
        self.diversities_sp = None
        self.scores_sp = None
        self.norm_scores_sp = None
        self.norm_diversities_sp = None
        if self._plot_rewards:
            self.rewards_sp = Histogram(title="Rewards")
        if self._plot_scores:
            self.scores_sp = Histogram(title="Scores")
        if self._plot_virtual_rewards:
            self.virtual_reward_sp = Histogram(title="Virtual Rewards")
        if self._plot_diversities:
            self.diversities_sp = Histogram(title="Diversities")
        if self._plot_norm_scores:
            self.norm_scores_sp = Histogram(title="Normalized Scores")
        if self._plot_norm_diversities:
            self.norm_diversities_sp = Histogram(title="Normalized Diversities")

    def send(self):
        if self._plot_rewards:
            self.rewards_sp.send(self.get("rewards"))
        if self._plot_scores:
            self.scores_sp.send(self.get("scores"))
        if self._plot_diversities:
            self.diversities_sp.send(self.get("diversities"))
        if self._plot_virtual_rewards:
            self.virtual_reward_sp.send(self.get("virtual_rewards"))
        if self._plot_norm_scores:
            scores = self.get("scores")
            scores = -1.0 * scores if self.swarm.minimize else scores
            self.norm_scores_sp.send(relativize(scores))
        if self._plot_norm_diversities:
            self.norm_diversities_sp.send(relativize(self.get("diversities")))

    def panel(self):
        histograms = []
        if self._plot_rewards:
            histograms.append(self.rewards_sp.plot)
        if self._plot_scores:
            histograms.append(self.scores_sp.plot)
        if self._plot_diversities:
            histograms.append(self.diversities_sp.plot)

        norm_histograms = []
        if self._plot_virtual_rewards:
            norm_histograms.append(self.virtual_reward_sp.plot)
        if self._plot_norm_scores:
            norm_histograms.append(self.norm_scores_sp.plot)
        if self._plot_norm_diversities:
            norm_histograms.append(self.norm_diversities_sp.plot)

        return panel.Column(panel.Row(*histograms), panel.Row(*norm_histograms))



from typing import Iterable, Optional, Tuple, Union

import networkx as nx

from fragile.callbacks.data_tracking import StoreInitAction
from fragile.callbacks.tree import HistoryTree
from fragile.core.api_classes import Callback, EnvironmentAPI, PolicyAPI, WalkersAPI
from fragile.core.swarm import Swarm
from fragile.core.typing import InputDict, StateData, StateDict, Tensor


class DummyFMCPolicy(PolicyAPI):
    def __init__(self, inner_swarm, **kwargs):
        self.inner_swarm = inner_swarm
        super(DummyFMCPolicy, self).__init__(**kwargs)

    def select_actions(self, **kwargs) -> Union[Tensor, StateData]:
        return {}

    def act(self, inplace: bool = True, **kwargs) -> Union[None, StateData]:
        if not inplace:
            return {}


class EnvSwarm(EnvironmentAPI):
    def __init__(self, swarm):
        self.inner_swarm = swarm
        super(EnvSwarm, self).__init__(
            swarm=None,
            action_shape=swarm.env.action_shape,
            action_dtype=swarm.env.action_dtype,
            observs_shape=swarm.env.observs_shape,
            observs_dtype=swarm.env.observs_dtype,
        )

    def __getattr__(self, item):
        return getattr(self.inner_swarm.env, item)

    @property
    def inputs(self) -> InputDict:
        return self.inner_swarm.env.inputs

    @property
    def outputs(self) -> Tuple[str, ...]:
        return self.inner_swarm.env.outputs + ("scores",)

    @property
    def param_dict(self) -> StateDict:
        pdict = self.inner_swarm.env.param_dict
        return {**{"scores": pdict["rewards"]}, **pdict}

    def make_transitions(self, inplace: bool = True, **kwargs) -> Union[None, StateData]:
        env_data = self.swarm.inputs_from_swarm("env")
        env_data = self.inner_swarm.env.make_transitions(inplace=False, **env_data)
        env_data["scores"] = env_data["rewards"]
        if self.inner_swarm.walkers.accumulate_reward:
            env_data["scores"] += self.get("scores")
        if inplace:
            self.swarm.state.update(**env_data)
        else:
            return env_data

    def step(self, **kwargs) -> StateData:
        pass


class JumpToBestTree(HistoryTree):
    def __len__(self):
        return len(self.graph.nodes)

    def after_evolve(self):
        pass

    def after_env(self):
        pass

    def update_tree(self):
        pass


class JumpToBestEnv(EnvSwarm):
    def make_transitions(self, inplace: bool = True, **kwargs) -> Union[None, StateData]:
        return
        self.swarm.state.import_walker(self.inner_swarm.root.data)
        if hasattr(self.swarm, "root"):
            self.swarm.root.update_root()

        if not inplace:
            return {}

    def reset(
        self,
        inplace: bool = True,
        root_walker: Optional[StateData] = None,
        states: Optional[StateData] = None,
        **kwargs,
    ) -> Union[None, StateData]:
        return super(JumpToBestEnv, self).make_transitions()


class WalkersSwarm(WalkersAPI):
    def __init__(self, inner_swarm, **kwargs):
        self.inner_swarm = inner_swarm
        self.accumulate_reward = self.inner_swarm.walkers.accumulate_reward
        super(WalkersSwarm, self).__init__(**kwargs)

    @property
    def param_dict(self) -> StateDict:
        return self.inner_swarm.param_dict

    def __getattr__(self, item):
        return getattr(self.inner_swarm, item)

    def run_epoch(self, inplace: bool = True, **kwargs) -> StateData:
        self.inner_swarm.run(root_walker=dict(self.swarm.state))
        return {}

    def reset(self, inplace: bool = True, **kwargs):
        super(WalkersSwarm, self).reset(inplace=inplace, **kwargs)
        self.inner_swarm.reset()
        walker = self.inner_swarm.state.export_walker(0)
        self.swarm.state.update(**walker)


class JumpToBest(WalkersSwarm):
    def run_epoch(self, inplace: bool = True, **kwargs) -> StateData:

        super(JumpToBest, self).run_epoch(inplace=inplace, **kwargs)
        print(
            self.swarm.root.id_walker,
            self.inner_swarm.root.id_walker,
            self.inner_swarm.tree.data_tree.root_id,
        )
        root_graph = self.inner_swarm.tree.get_root_graph()
        self.swarm.tree.data_tree.compose(root_graph)
        self.swarm.state.update(**self.inner_swarm.root.data)
        self.swarm.root.update_root()
        print(
            "self tree",
            nx.is_tree(self.tree.graph),
            "inner",
            nx.is_tree(self.inner_swarm.tree.graph),
            "root",
            nx.is_tree(root_graph),
        )
        return {}

    def reset(self, inplace: bool = True, **kwargs):
        super(JumpToBest, self).reset(inplace=inplace, **kwargs)
        root_graph = self.inner_swarm.tree.get_root_graph()
        print(
            "self tree",
            nx.is_tree(self.tree.graph),
            "inner",
            nx.is_tree(self.inner_swarm.tree.graph),
            "root",
            nx.is_tree(root_graph),
        )
        self.swarm.tree.data_tree._data = root_graph
        walker = self.inner_swarm.state.export_walker(0)
        self.swarm.state.update(**walker)
        self.swarm.root.reset()
        self.swarm.root.update_root()


class FMCSwarm(Swarm):
    walkers_last = False

    def __init__(
        self,
        swarm: Swarm,
        policy: PolicyAPI = None,
        env: EnvironmentAPI = None,
        callbacks: Optional[Iterable[Callback]] = None,
        minimize: bool = False,
        max_epochs: int = 1e100,
    ):
        swarm.register_callback(StoreInitAction())
        # if hasattr(swarm.policy, "bounds"):
        #    swarm.register_callback(BoundaryInitAction())
        self._swarm = swarm
        policy = DummyFMCPolicy(swarm) if policy is None else policy
        env = EnvSwarm(swarm) if env is None else env
        super(FMCSwarm, self).__init__(
            n_walkers=swarm.n_walkers,
            policy=policy,
            env=env,
            callbacks=callbacks,
            minimize=minimize,
            max_epochs=int(max_epochs),
            walkers=JumpToBest(swarm),
        )

    @property
    def swarm(self) -> Swarm:
        return self._swarm

    @property
    def n_walkers(self) -> int:
        return 1

    def inputs_from_swarm(self, name):
        def get_names():
            names = []
            inputs = getattr(self.swarm, name).inputs
            for k, v in inputs.items():
                is_optional = v.get("optional", False)
                if not is_optional or (is_optional and k in self.swarm.param_dict.keys()):
                    names.append(k)
            return names

        return {name: self.state[name] for name in get_names()}



import judo
from judo.functions.random import random_state
import pytest

from fragile.algorithms.fmc import FMCSwarm
from fragile.callbacks.root_walker import BestWalker
from fragile.callbacks.tree import HistoryTree
from fragile.core.policy import Gaussian
from fragile.core.swarm import FunctionMapper
from fragile.core.walkers import Walkers
from fragile.optimize.benchmarks import Rastrigin
from fragile.optimize.env import MinimizerWrapper


def create_rastrigin(max_epochs=500):
    walkers = Walkers(score_scale=2, accumulate_reward=False)
    env = Rastrigin(2)
    policy = Gaussian(scale=0.25, second_order=False)
    callbacks = [BestWalker()]
    n_walkers = 32
    swarm = FunctionMapper(
        policy=policy,
        walkers=walkers,
        env=env,
        callbacks=callbacks,
        n_walkers=n_walkers,
        max_epochs=max_epochs,
    )
    return swarm


def create_minimzer_rastrigin(max_epochs=500):
    walkers = Walkers(score_scale=2, accumulate_reward=False)
    env = MinimizerWrapper(Rastrigin(2))
    policy = Gaussian(scale=0.25, second_order=False)
    callbacks = [BestWalker()]
    n_walkers = 32
    swarm = FunctionMapper(
        policy=policy,
        walkers=walkers,
        env=env,
        callbacks=callbacks,
        n_walkers=n_walkers,
        max_epochs=max_epochs,
    )
    return swarm


def rastrigin_fmc():
    swarm = create_rastrigin(max_epochs=10)
    fmc = FMCSwarm(
        minimize=True,
        swarm=swarm,
        max_epochs=150,
        callbacks=[BestWalker(always_update=True), HistoryTree(names=["states", "actions"])],
    )
    return fmc


swarms = [(create_rastrigin, 1), (rastrigin_fmc, 18), (create_minimzer_rastrigin, 1)]


@pytest.mark.parametrize("swarm_and_target", swarms)
def test_score_gets_higher(swarm_and_target):
    random_state.seed(160290)
    swarm, target_score = swarm_and_target
    swarm = swarm()
    swarm.run()
    root_score = float(swarm.root.score)
    beats_target = root_score <= target_score if swarm.minimize else root_score >= target_score
    assert beats_target, (root_score, target_score)



from functools import partial

import numpy as np
import numpy.linalg

from fragile.core.fractalai import relativize


def numerical_grad_single_vector(theta, f, dx=1e-3, order=1):
    """return numerical estimate of the local gradient
    The gradient is computer by using the Taylor expansion approximation over
    each dimension:
        function(t + dt) = function(t) + h df/dt(t) + h^2/2 d^2f/dt^2 + ...
    The first order gives then:
        df/dt = (function(t + dt) - function(t)) / dt + O(dt)
    Note that we could also compute the backwards version by subtracting dt instead:
        df/dt = (function(t) - function(t -dt)) / dt + O(dt)
    A better approach is to use a 3-step formula where we evaluate the
    derivative on both sides of a chosen point t using the above forward and
    backward two-step formulae and taking the average afterward. We need to use the Taylor
     expansion to higher order:
        function (t +/- dt) = function (t) +/- dt df/dt + dt ^ 2 / 2  dt^2 function/dt^2 +/- \
        dt ^ 3 d^3 function/dt^3 + O(dt ^ 4)
        df/dt = (function(t + dt) - function(t - dt)) / (2 * dt) + O(dt ^ 3)
    Note: that the error is now of the order of dt ^ 3 instead of dt
    In a same manner we can obtain the next order by using function(t +/- 2 * dt):
        df/dt = (function(t - 2 * dt) - 8 function(t - dt)) + 8 function(t + dt) - \
        function(t + 2 * dt) / (12 * dt) + O(dt ^ 4)
    In the 2nd order, two additional function evaluations are required (per dimension), implying a
    more time-consuming algorithm. However, the approximation error is of the order of dt ^ 4

    INPUTS
    ------
    theta: ndarray[float, ndim=1]
        vector value around which estimating the gradient
    function: callable
        function from which estimating the gradient
    KEYWORDS
    --------
    dx: float
        pertubation to apply in each direction during the gradient estimation
    order: int in [1, 2]
        order of the estimates:
            1 uses the central average over 2 points
            2 uses the central average over 4 points
    OUTPUTS
    -------
    df: ndarray[float, ndim=1]
        gradient vector estimated at theta
    COST: the gradient estimation need to evaluates ndim * (2 * order) points (see above)
    CAUTION: if dt is very small, the limited numerical precision can result in big errors.
    """
    ndim = len(theta)
    df = np.empty(ndim, dtype=float)
    if order == 1:
        cst = 0.5 / dx
        for k in range(ndim):
            dt = np.zeros(ndim, dtype=float)
            dt[k] = dx
            df[k] = (f(theta + dt) - f(theta - dt)) * cst
    elif order == 2:
        cst = 1.0 / (12.0 * dx)
        for k in range(ndim):
            dt = np.zeros(ndim, dtype=float)
            dt[k] = dx
            df[k] = cst * (
                f(theta - 2 * dt) - 8.0 * f(theta - dt) + 8.0 * f(theta + dt) - f(theta + 2.0 * dt)
            )
    return df


def numerical_grad(theta, f, dx=1e-3, order=1):
    """return numerical estimate of the local gradient
    The gradient is computer by using the Taylor expansion approximation over
    each dimension:
        function(t + dt) = function(t) + h df/dt(t) + h^2/2 d^2f/dt^2 + ...
    The first order gives then:
        df/dt = (function(t + dt) - function(t)) / dt + O(dt)
    Note that we could also compute the backwards version by subtracting dt instead:
        df/dt = (function(t) - function(t -dt)) / dt + O(dt)
    A better approach is to use a 3-step formula where we evaluate the
    derivative on both sides of a chosen point t using the above forward and
    backward two-step formulae and taking the average afterward. We need to use the
     Taylor expansion to higher order:
    function (t +/- dt) = function (t) +/- dt df/dt + dt ^ 2 / 2  dt^2
    function/dt^2 +/- dt ^ 3 d^3 function/dt^3 + O(dt ^ 4)
        df/dt = (function(t + dt) - function(t - dt)) / (2 * dt) + O(dt ^ 3)
    Note: that the error is now of the order of dt ^ 3 instead of dt
    In a same manner we can obtain the next order by using function(t +/- 2 * dt):
        df/dt = (function(t - 2 * dt) - 8 function(t - dt)) + 8 function(t + dt) - \
        function(t + 2 * dt) / (12 * dt) + O(dt ^ 4)
    In the 2nd order, two additional function evaluations are required (per dimension), implying a
    more time-consuming algorithm. However the approximation error is of the order of dt ^ 4
    INPUTS
    ------
    theta: ndarray[float, ndim=1]
        vector value around which estimating the gradient
    function: callable
        function from which estimating the gradient
    KEYWORDS
    --------
    dx: float
        pertubation to apply in each direction during the gradient estimation
    order: int in [1, 2]
        order of the estimates:
            1 uses the central average over 2 points
            2 uses the central average over 4 points
    OUTPUTS
    -------
    df: ndarray[float, ndim=1]
        gradient vector estimated at theta
    COST: the gradient estimation need to evaluates ndim * (2 * order) points (see above)
    CAUTION: if dt is very small, the limited numerical precision can result in big errors.
    """
    ndim = theta.shape[1]
    df = np.empty_like(theta)
    if order == 1:
        cst = 0.5 / dx
        for k in range(ndim):
            dt = np.zeros(theta.shape, dtype=float)
            dt[:, k] = dx
            df[:, k] = (f(theta + dt) - f(theta - dt)) * cst
    elif order == 2:
        cst = 1.0 / (12.0 * dx)
        for k in range(ndim):
            dt = np.zeros(theta.shape, dtype=float)
            dt[:, k] = dx
            df[:, k] = cst * (
                f(theta - 2 * dt) - 8.0 * f(theta - dt) + 8.0 * f(theta + dt) - f(theta + 2.0 * dt)
            )
    return df


def leapfrog(init_pos, init_momentum, grad, step_size, function, force=None):
    """Perfom a leapfrog jump in the Hamiltonian space
    INPUTS
    ------
    init_pos: ndarray[float, ndim=1]
        initial parameter position
    init_momentum: ndarray[float, ndim=1]
        initial momentum
    grad: float
        initial gradient value
    step_size: float
        step size
    function: callable
        it should return the log probability and gradient evaluated at theta
        logp, grad = function(theta)
    OUTPUTS
    -------
    new_position: ndarray[float, ndim=1]
        new parameter position
    new_momentum: ndarray[float, ndim=1]
        new momentum
    gradprime: float
        new gradient
    new_potential: float
        new lnp
    """
    force = 0.0 if force is None else force
    # make half step in init_momentum
    new_momentum = init_momentum + 0.5 * step_size * (force - grad)
    # make new step in theta
    new_position = init_pos + step_size * new_momentum
    # compute new gradient
    new_potential, gradprime = function(new_position)
    # make half step in init_momentum again
    new_momentum = new_momentum + 0.5 * step_size * (force - gradprime)
    return new_position, new_momentum, gradprime, new_potential


def fai_leapfrog(init_pos, init_momentum, grad, step_size, function, force=None):
    """Perfom a leapfrog jump in the Hamiltonian space
    INPUTS
    ------
    init_pos: ndarray[float, ndim=1]
        initial parameter position
    init_momentum: ndarray[float, ndim=1]
        initial momentum
    grad: float
        initial gradient value
    step_size: float
        step size
    function: callable
        it should return the log probability and gradient evaluated at theta
        logp, grad = function(theta)
    OUTPUTS
    -------
    new_position: ndarray[float, ndim=1]
        new parameter position
    new_momentum: ndarray[float, ndim=1]
        new momentum
    gradprime: float
        new gradient
    new_potential: float
        new lnp
    """
    force = 0.0 if force is None else force
    potential, grad = function(init_pos)
    pot = relativize(potential).reshape(-1, 1)
    grad = (grad / numpy.linalg.norm(grad, axis=1).reshape(-1, 1)) * pot
    # make half step in init_momentum
    new_momentum = init_momentum + 0.5 * step_size * (force - grad)
    # make new step in theta
    new_position = init_pos + step_size * new_momentum
    # compute new gradient
    new_potential, gradprime = function(new_position)
    new_pot = relativize(new_potential).reshape(-1, 1)
    gradprime = (gradprime / numpy.linalg.norm(gradprime, axis=1).reshape(-1, 1)) * new_pot
    # make half step in init_momentum again
    new_momentum = new_momentum + 0.5 * step_size * (force - gradprime)
    return new_position, new_momentum, gradprime, new_potential


class GradFuncWrapper:
    """Create a function-like object that combines provided lnp and grad(lnp)
    functions into one as required by nuts6.
    Both functions are stored as partial function allowing to fix arguments if
    the gradient function is not provided, numerical gradient will be computed
    By default, arguments are assumed identical for the gradient and the
    likelihood. However you can change this behavior using set_xxxx_args.
    (keywords are also supported)
    if verbose property is set, each call will print the log-likelihood value
    and the theta point
    """

    def __init__(self, function, grad_func=None, dx=1e-3, order=1):
        self.function = function
        if grad_func is not None:
            self.grad_func = grad_func
        else:
            self.grad_func = partial(numerical_grad, f=self.function, dx=dx, order=order)
        self.verbose = False

    def __call__(self, theta):
        r = (self.function(theta), self.grad_func(theta))
        if self.verbose:
            print(r[0], theta)
        return r



import tempfile

import panel
from PIL import Image


def create_gif(data, filename=None, fps=10, optimize=False):
    duration = int((len(data) / fps) * 20)
    filename = tempfile.NamedTemporaryFile(suffix="a.gif") if filename is None else filename
    images = [Image.fromarray(v) for v in data]
    images[0].save(
        filename,
        save_all=True,
        append_images=images[1:],
        optimize=optimize,
        duration=duration,
        loop=0,
    )
    return filename


def show_root_game(swarm, fps=10, optimize=False):
    vals, *_ = zip(*list(swarm.tree.iterate_root_path(names=["rgb"])))
    f = create_gif(vals, fps=fps, optimize=optimize)
    return panel.pane.GIF(f)



from importlib.machinery import SourceFileLoader
from pathlib import Path

from setuptools import find_packages, setup


version = SourceFileLoader(
    "fragile.version",
    str(Path(__file__).parent / "fragile" / "version.py"),
).load_module()

with open(Path(__file__).with_name("README.md"), encoding="utf-8") as f:
    long_description = f.read()

# Module-specific dependencies.
extras = {
    "atari": ["atari-py==0.1.1", "opencv-python", "gym==0.17.3", "pillow-simd", "plangym>=0.0.31"],
    "dataviz": [
        "matplotlib==3.7.1",
        "bokeh==2.4.0",
        "pandas==1.5.3",
        "panel==0.14.4",
        "holoviews",
        "hvplot",
        "plotly",
        "streamz",
        "param",
        "selenium",
        "pyarrow",
    ],
    "test": [
        "pytest==6.2.5",
        "pytest-cov==3.0.0",
        "pytest-xdist==2.4.0",
        "pytest-rerunfailures==10.2",
        "hypothesis==6.24.6",
    ],
    "ray": ["ray>=1.0.1.post1", "setproctitle"],
}

# Meta dependency groups.
extras["all"] = [item for group in extras.values() for item in group]

setup(
    name="fragile",
    description="Framework for developing FractalAI based algorithms.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    packages=find_packages(),
    version=version.__version__,
    license="MIT",
    author="Guillem Duran Ballester",
    author_email="info@fragile.tech",
    url="https://github.com/FragileTech/fragile",
    download_url="https://github.com/FragileTech/fragile",
    keywords=["reinforcement learning", "artificial intelligence", "monte carlo", "planning"],
    tests_require=["pytest>=5.3.5", "hypothesis>=5.6.0"],
    extras_require=extras,
    install_requires=[
        "einops",
        "flogging",
        "judo>=0.0.15",
        "networkx",
        "numba",
        "scipy",
        "plangym>=0.0.31",
        "tqdm",
    ],
    package_data={"": ["README.md"]},
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Environment :: Console",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: MIT License",
        "Operating System :: POSIX",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Topic :: Software Development :: Libraries",
    ],
)



import judo

from fragile.core.api_classes import Callback


class KillWorstWalkers(Callback):
    name = "oobs_kill_worst"
    default_inputs = {"oobs": {}, "scores": {}}

    def __init__(self, stds: float = 1.0, **kwargs):
        super(KillWorstWalkers, self).__init__(**kwargs)
        self._stds = stds

    def oob_condition(self):
        scores = self.get("scores")
        if self.swarm.minimize:
            return scores > (scores.min() + scores.std() * self._stds)
        else:
            return scores < (scores.max() - scores.std() * self._stds)

    def after_env(self):
        oobs = judo.logical_or(self.get("oobs"), self.oob_condition())
        self.update(oobs=oobs)



import judo
import numpy as np
import pytest

from fragile.core.state import State


@pytest.fixture(scope="module")
def state():
    s = State(n_walkers=3, param_dict={"x": {"shape": None, "dtype": int}})
    s.reset()
    return s


def test_values(state):
    vals = list(state.values())
    assert np.array_equal(vals, [[None, None, None]]), vals


def test_items(state):
    param_dict = dict(state.param_dict)
    param_dict["y"] = {"shape": None, "dtype": int}
    state = State(n_walkers=state.n_walkers, param_dict=param_dict)
    state.update(x=[1, 2, 3], y=[4, 5, 6])
    my_value = np.array(list(state.items()), dtype=object)
    test_value = np.array([("x", [1, 2, 3]), ("y", [4, 5, 6])], dtype=object)
    assert np.array_equal(my_value, test_value)


def test_itervals(state):
    param_dict = dict(state.param_dict)
    param_dict["y"] = {"shape": None, "dtype": int}
    state = State(n_walkers=state.n_walkers, param_dict=param_dict)
    state.update(x=[1, 2, 3], y=[4, 5, 6])
    assert np.array_equal(list(state.itervals()), [(1, 4), (2, 5), (3, 6)])


def test_iteritems(state):
    param_dict = dict(state.param_dict)
    param_dict["y"] = {"shape": None, "dtype": int}
    state = State(n_walkers=state.n_walkers, param_dict=param_dict)
    state.update(x=[1, 2, 3], y=[4, 5, 6])
    assert np.array_equal(
        list(state.iteritems()),
        [(("x", "y"), (1, 4)), (("x", "y"), (2, 5)), (("x", "y"), (3, 6))],
    )


def test_update(state):
    param_dict = dict(state.param_dict)
    param_dict["y"] = {"shape": None, "dtype": int}
    state = State(n_walkers=state.n_walkers, param_dict=param_dict)
    state.update(x=[1, 2, 3], y=[4, 5, 6])
    assert np.array_equal(state["x"], [1, 2, 3])
    assert np.array_equal(state["y"], [4, 5, 6])


def test_to_dict(state):
    state["x"] = [1, 2, 3]
    state_dict = state.to_dict()
    assert isinstance(state_dict, dict)
    assert "x" in state_dict.keys()
    assert "n_walkers" not in state_dict.keys()
    assert state.n_walkers == len(state_dict["x"])
    assert "param_dict" not in state_dict.keys()
    assert np.array_equal(state_dict["x"], [1, 2, 3])


def test_copy(state):
    state["x"] = [1, 2, 3]
    state_copy = state.copy()
    assert isinstance(state_copy, State)
    assert np.array_equal(list(state_copy.values()), [[1, 2, 3]])
    assert np.array_equal(list(state_copy["x"]), [1, 2, 3])


def test_reset(state):
    state["x"] = [1, 2, 3]
    state.reset()
    assert np.array_equal(list(state.values()), [[None, None, None]])


def test_params_to_arrays(state):
    param_dict_list = {
        "list_param_int": {"shape": None, "dtype": int},
        "list_param_list": {"shape": None, "dtype": list},
        "list_param_dict": {"shape": None, "dtype": dict},
        "list_param_set": {"shape": None, "dtype": set},
    }
    tensor_dict = state.params_to_arrays(param_dict_list, n_walkers=2)
    assert isinstance(tensor_dict, dict)
    assert "list_param_int" in tensor_dict.keys()
    assert "list_param_list" in tensor_dict.keys()
    assert "list_param_dict" in tensor_dict.keys()
    assert "list_param_set" in tensor_dict.keys()
    assert isinstance(tensor_dict["list_param_int"], list)
    assert isinstance(tensor_dict["list_param_list"], list)
    assert isinstance(tensor_dict["list_param_dict"], list)
    assert isinstance(tensor_dict["list_param_set"], list)
    assert np.array_equal(tensor_dict["list_param_int"], [None, None])
    assert np.array_equal(tensor_dict["list_param_list"], [None, None])
    assert np.array_equal(tensor_dict["list_param_dict"], [None, None])



from typing import Any, Dict

import judo
from judo.data_types import dtype
import numpy as np
import pytest

from fragile.core.state import State


param_dict_list = {
    "list_param_int": {"shape": None, "dtype": int},
    "list_param_list": {"shape": None, "dtype": list},
    "list_param_dict": {"shape": None, "dtype": dict},
    "list_param_set": {"shape": None, "dtype": set},
}

param_dict_vector = {
    "vec_float_16": {"dtype": dtype.float16},
    "vec_float_32": {"dtype": dtype.float32},
    "vec_bool": {"dtype": dtype.bool},
    "vec_int_shape": {"shape": tuple(), "dtype": dtype.int64},
}

param_dict_tensor = {
    "tensor_float_20_17": {"shape": (20, 17), "dtype": dtype.float16},
    "tensor_float_1": {"shape": (1,), "dtype": dtype.float32},
    "tensor_int_1_17_5": {"shape": (1, 17, 5), "dtype": dtype.int64},
}


param_dict_mix = {
    "observs": {"shape": (20, 17), "dtype": dtype.float16},
    "rewards": {"dtype": dtype.float32},
    "oobs": {"dtype": dtype.bool},
    "actions": {"shape": (1, 17), "dtype": dtype.int64},
    "list_param": {"shape": None, "dtype": int},
}

param_dict_examples = [param_dict_list, param_dict_vector, param_dict_tensor, param_dict_mix]
param_dict_ids = ["list_param_dict", "vector_param_dict", "tensor_param_dict", "mix_param_dict"]


@pytest.fixture(params=param_dict_examples, ids=param_dict_ids)
def param_dict(request) -> Dict[str, Any]:
    return request.param


@pytest.fixture()
def n_walkers() -> int:
    return 7


class TestState:
    def test_init(self, param_dict, n_walkers):
        state = State(n_walkers=n_walkers, param_dict=param_dict)
        state.reset()
        for name in param_dict.keys():
            assert name in state.names

    def test_getitem(self, param_dict, n_walkers):
        state = State(n_walkers=n_walkers, param_dict=param_dict)
        state.reset()
        for name in param_dict.keys():
            assert hasattr(state, name)
            assert state[name] is not None
            assert np.array_equal(judo.to_numpy(state[name]), judo.to_numpy(state.get(name)))

    def test_setitem(self, param_dict, n_walkers):
        state = State(n_walkers=n_walkers, param_dict=param_dict)
        state.reset()
        for name in param_dict.keys():
            add_val = [1] if isinstance(state[name], list) else 1
            new_value = state[name] + add_val
            state[name] = new_value
            assert np.array_equal(judo.to_numpy(state[name]), judo.to_numpy(new_value))

    def test_reset(self, param_dict, n_walkers):
        state = State(n_walkers=n_walkers, param_dict=param_dict)
        state.reset()
        for name, value in param_dict.items():
            assert len(state[name]) == n_walkers
            if "dtype" in value and not isinstance(state[name], list):
                assert state[name].dtype == value["dtype"]
            if "shape" in value and value["shape"] is not None:
                assert state[name].shape == (n_walkers,) + value["shape"]
            elif value.get("shape", False) is not None:
                assert state[name].shape == (n_walkers,)

    def __test_setitem_wrong_shape(self, param_dict, n_walkers):
        state = State(n_walkers=n_walkers, param_dict=param_dict)
        state.reset()
        for name in param_dict.keys():
            if "shape" in param_dict[name]:
                with pytest.raises(ValueError):
                    state[name] = [1] * (len(param_dict[name]["shape"]) + 1)

    def __test_setitem_wrong_dtype(self, param_dict, n_walkers):
        state = State(n_walkers=n_walkers, param_dict=param_dict)
        state.reset()
        for name in param_dict.keys():
            if "dtype" in param_dict[name]:
                with pytest.raises(TypeError):
                    state[name] = "wrong_dtype"

    def test_copy(self, param_dict, n_walkers):
        state = State(n_walkers=n_walkers, param_dict=param_dict)
        state.reset()
        copy_state = state.copy()
        assert copy_state is not state
        assert copy_state.names == state.names
        for name in param_dict.keys():
            assert np.array_equal(judo.to_numpy(copy_state[name]), judo.to_numpy(state[name]))

    def test_copy_change(self, param_dict, n_walkers):
        state = State(n_walkers=n_walkers, param_dict=param_dict)
        state.reset()
        copy_state = state.copy()
        for name in param_dict.keys():
            if isinstance(state[name], (int, float, bool)):
                copy_state[name] += 1
                assert copy_state[name] != state[name]
            elif isinstance(state[name], (list, tuple, set)):
                copy_state[name].append(1)
                assert len(copy_state[name]) != len(state[name])
            elif isinstance(state[name], dict):
                copy_state[name]["new_key"] = "new_value"
                assert len(copy_state[name]) != len(state[name])

    def test_len(self, param_dict, n_walkers):
        state = State(n_walkers=n_walkers, param_dict=param_dict)
        state.reset()
        n_walkers = len(state)
        assert n_walkers == state.n_walkers



import os

import einops
from evotorch.algorithms import GeneticAlgorithm
from evotorch.algorithms.searchalgorithm import SearchAlgorithm
from evotorch.core import Problem, SolutionBatch
from evotorch.operators import GaussianMutation, OnePointCrossOver
import judo
import pytest

from fragile.callbacks.root_walker import BestWalker
from fragile.core.policy import ZeroContinuous
from fragile.core.swarm import FunctionMapper
from fragile.core.walkers import Walkers
from fragile.optimize.evotorch import EvotorchEnv


def create_searcher(popsize):
    import torch

    def sphere(x: torch.Tensor) -> torch.Tensor:
        return torch.sum(x.pow(2.0))

    problem = Problem(
        "min",
        sphere,
        initial_bounds=(([-5.12, -5.12]), ([5.12, 5.12])),
        solution_length=2,
        dtype=judo.dtype.float32,
        eval_dtype=judo.dtype.float32,
        device=judo.Backend.get_device(),
    )
    searcher = GeneticAlgorithm(
        problem,
        popsize=popsize,
        re_evaluate=True,
        elitist=True,
        operators=[
            OnePointCrossOver(problem, tournament_size=3),
            GaussianMutation(problem, stdev=0.05),
        ],
    )
    return searcher


def create_evotorch(max_epochs=10):
    n_walkers = 200
    walkers = Walkers(score_scale=2, accumulate_reward=False, freeze_walkers=False)
    searcher = create_searcher(popsize=n_walkers)
    env = EvotorchEnv(searcher)
    policy = ZeroContinuous()
    callbacks = [BestWalker()]

    swarm = FunctionMapper(
        policy=policy,
        walkers=walkers,
        env=env,
        callbacks=callbacks,
        n_walkers=n_walkers,
        max_epochs=max_epochs,
    )
    return swarm


@pytest.fixture(scope="module")
def swarm():
    swarm = create_evotorch(max_epochs=200)
    return swarm


class TestEvotorchEnv:
    def test_properties(self, swarm):
        assert isinstance(swarm.env.algorithm, SearchAlgorithm)
        assert isinstance(swarm.env.population, SolutionBatch)
        assert isinstance(swarm.env.problem, Problem)
        assert swarm.env.population == swarm.env.algorithm.population
        assert swarm.env.problem == swarm.env.algorithm.problem
        assert swarm.env.dtype == swarm.env.algorithm.population.dtype
        assert swarm.env.eval_dtype == swarm.env.algorithm.population.eval_dtype
        assert swarm.env.solution_length == swarm.env.algorithm.population.solution_length

    def test_score_gets_higher(self, swarm):
        target_score = 0.005
        swarm.run()
        root_score = float(swarm.root.score)
        beats_target = root_score <= target_score if swarm.minimize else root_score >= target_score
        assert beats_target, (root_score, target_score)

    def test_get_bounds(self, swarm):
        bounds = swarm.env._get_bounds()
        assert isinstance(bounds, judo.Bounds)
        assert (bounds.low == einops.asnumpy(swarm.env.problem.initial_lower_bounds)).all()
        assert (bounds.high == einops.asnumpy(swarm.env.problem.initial_upper_bounds)).all()



import copy
from typing import Any, Dict, Generator, List, Optional, Set, Tuple, Union

import judo
from judo import tensor
from judo.functions.api import API
from judo.functions.hashing import hasher

from fragile.core.typing import StateDict, Tensor


class State:
    """
    Data structure that handles the data defining a population of walkers.

    Each population attribute will be stored as a tensor with its first dimension
    (batch size) representing each walker.

    In order to define a tensor attribute, a `param_dict` dictionary needs to be
    specified using the following structure:

    Example:
        >>> attr_dict = {'name': {'shape': Optional[tuple|None], 'dtype': dtype},  # doctest: +SKIP
        ...             'biases': {'shape': (10,), 'dtype': float},
        ...             'vector': {'dtype': 'float32'},
        ...             'sequence': {'shape': None, 'dtype': 'float32'}
        ...             }

    Where tuple is a tuple indicating the shape of the desired tensor. The created
    arrays will be accessible through the `name_1` attribute of the class, or by
    indexing the class with `states["name_1"]`.

    If `size` is not defined, the attribute will be considered a vector of length
    `n_walkers`.


    Args:
        n_walkers (int): The number of items in the first dimension of the tensors.
        param_dict (StateDict): Dictionary defining the attributes of the tensors.

    Attributes:
        n_walkers (int): The number of walkers that this instance represents.
        param_dict (StateDict): A dictionary containing the shape and type of each
            walker's attribute.
        names (Tuple[str]): The name of the walker's attributes tracked by this
            instance.
        tensor_names (Set[str]): The name of the walker's attributes that correspond
            to tensors.
        list_names (Set[str]): The name of the walker's attributes that correspond
            to lists of objects.
        vector_names (Set[str]): The name of the walker's attributes that correspond
            to vectors of scalars.
    """

    def __init__(self, n_walkers: int, param_dict: StateDict):
        """
        Initialize a :class:`SwarmState`.

        Args:
             n_walkers: The number of items in the first dimension of the tensors.
             param_dict: Dictionary defining the attributes of the tensors.

        """

        def shape_is_vector(v):
            shape = v.get("shape", ())
            return not (shape is None or (isinstance(shape, tuple) and len(shape) > 0))

        self._param_dict = param_dict
        self._n_walkers = n_walkers
        self._names = tuple(param_dict.keys())
        self._list_names = set(k for k, v in param_dict.items() if v.get("shape", 1) is None)
        self._tensor_names = set(k for k in self.names if k not in self._list_names)
        self._vector_names = set(k for k, v in param_dict.items() if shape_is_vector(v))

    @property
    def n_walkers(self) -> int:
        """Return the number of walkers that this instance represents."""
        return self._n_walkers

    @property
    def param_dict(self) -> StateDict:
        """Return a dictionary containing the shape and type of each walker's attribute."""
        return self._param_dict

    @property
    def names(self) -> Tuple[str]:
        """Return the name of the walker's attributes tracked by this instance."""
        return self._names

    @property
    def tensor_names(self) -> Set[str]:
        """Return the name of the walker's attributes that correspond to tensors."""
        return self._tensor_names

    @property
    def list_names(self) -> Set[str]:
        """Return the name of the walker's attributes that correspond to lists of objects."""
        return self._list_names

    @property
    def vector_names(self) -> Set[str]:
        """Return the name of the walker's attributes that correspond to vectors of scalars."""
        return self._vector_names

    def __len__(self) -> int:
        """Length is equal to n_walkers."""
        return self._n_walkers

    def __setitem__(self, key, value: Union[Tuple, List, Tensor]):
        """
        Allow the class to set its attributes as if it was a dict.

        Args:
            key: Attribute to be set.
            value: Value of the target attribute.

        Returns:
            None.

        """
        setattr(self, key, value)

    def __getitem__(self, item: str) -> Union[Tensor, List[Tensor], "SwarmState"]:
        """
        Query an attribute of the class as if it was a dictionary.

        Args:
            item: Name of the attribute to be selected.

        Returns:
            The corresponding item.

        """
        return getattr(self, item)

    def __repr__(self) -> str:
        """Return a string that provides a nice representation of this instance attributes."""
        string = f"{self.__class__.__name__} with {self._n_walkers} walkers\n"
        for k, v in self.items():
            shape = v.shape if hasattr(v, "shape") else None
            new_str = "{}: {} {}\n".format(k, v.__class__.__name__, shape)
            string += new_str
        return string

    def __hash__(self) -> int:
        """Return an integer that represents the hash of the current instance."""
        return hasher.hash_state(self)

    def hash_attribute(self, name: str) -> int:
        """Return a unique id for a given attribute."""
        return hasher.hash_tensor(self[name])

    def hash_batch(self, name: str) -> List[int]:
        """Return a unique id for each walker attribute."""
        return hasher.hash_iterable(self[name])

    def get(self, name: str, default=None, raise_error: bool = True):
        """
        Get an attribute by key and return the default value if it does not exist.

        Args:
            name: Attribute to be recovered.
            default: Value returned in case the attribute is not part of state.
            raise_error: If True, raise AttributeError if name is not present in states.

        Returns:
            Target attribute if found in the instance, otherwise returns the
            default value.

        """
        if name not in self.names:
            if raise_error:
                raise AttributeError(f"{name} not present in states.")
            return default
        return self[name]

    def keys(self) -> "_dict_keys[str, Dict[str, Any]]":  # pyflakes: disable=F821
        """Return a generator for the values of the stored data."""
        return self.param_dict.keys()

    def values(self) -> Generator:
        """Return a generator for the values of the stored data."""
        return (self[name] for name in self.keys())

    def items(self) -> Generator:
        """Return a generator for the attribute names and the values of the stored data."""
        return ((name, self[name]) for name in self.keys())

    def itervals(self):
        """
        Iterate the states attributes by walker.

        Returns:
            Tuple containing all the names of the attributes, and the values that
            correspond to a given walker.

        """
        if len(self) <= 1:
            yield self.values()
            raise StopIteration
        for i in range(self.n_walkers):
            yield tuple(v[i] for v in self.values())

    def iteritems(self):
        """
        Iterate the states attributes by walker.

        Returns:
            Tuple containing all the names of the attributes, and the values that
            correspond to a given walker.

        """
        if self.n_walkers < 1:
            return self.values()
        for i in range(self.n_walkers):
            yield tuple(self.names), tuple(v[i] for v in self.values())

    def update(self, other: Union["SwarmState", Dict[str, Tensor]] = None, **kwargs) -> None:
        """
        Modify the data stored in this instance.

        Existing attributes will be updated, and no new attributes can be created.

        Args:
            other (Union[SwarmState, Dict[str, Tensor]], optional): Other SwarmState
                instance to copy upon update. Defaults to None.
            **kwargs: Extra key-value pairs of attributes to add to or modify the current state.

        Example:
            >>> s = State(2, {'name': {'shape': (3, 4), "dtype": bool}})
            >>> s.update({'name': np.ones((3, 4))})
            >>> len(s.names)
            1
            >>> s['name']
            array([[ True,  True,  True,  True],
                   [ True,  True,  True,  True],
                   [ True,  True,  True,  True]])
        """
        new_values = other.to_dict() if isinstance(other, SwarmState) else (other or {})
        new_values.update(kwargs)
        for name, val in new_values.items():
            val = self._parse_value(name, val)
            setattr(self, name, val)

    def to_dict(self) -> Dict[str, Union[Tensor, list]]:
        """Return the stored data as a dictionary of arrays."""
        return {name: value for name, value in self.items()}

    def copy(self) -> "SwarmState":
        """Crete a copy of the current instance."""
        new_states = self.__class__(n_walkers=self.n_walkers, param_dict=self.param_dict)
        new_states.update(**copy.deepcopy(dict(self)))
        return new_states

    def reset(self) -> None:
        """Reset the values of the class"""
        try:
            data = self.params_to_arrays(self.param_dict, self.n_walkers)
        except Exception as e:
            print(self.param_dict)
            raise e
        self.update(data)

    def params_to_arrays(self, param_dict: StateDict, n_walkers: int) -> Dict[str, Tensor]:
        """
        Create a dictionary containing arrays specified by ``param_dict``, \
        the attribute dictionary.

        This method creates a dictionary containing arrays specified by the attribute dictionary.
        The attribute dictionary defines the attributes of the tensors, and ``n_walkers`` is
        the number of items in the first dimension of the data tensors.

        The method returns a dictionary with the same keys as ``param_dict``, containing
        arrays specified by the values in the attribute dictionary. The method achieves this by
        iterating through each item in the attribute dictionary, creating a copy of the value, and
        checking if the shape of the value is specified.

        If the shape is specified, the method initializes the tensor to zeros using the
        shape and dtype specified in the attribute dictionary. If the shape is not specified
        or if the key is in ``self.list_names``, the method initializes the value as a list of
        `None` with ``n_walkers`` items. If the key is ``'label'``, the method initializes
        the value as a list of empty strings with ``n_walkers`` items.

        Args:
            param_dict: Dictionary defining the attributes of the tensors.
            n_walkers: Number of items in the first dimension of the data tensors.

        Returns:
            Dictionary with the same names as the attribute dictionary, containing arrays specified
            by the values in the attribute dictionary.

        Example:
            >>> attr_dict = {'weights': {'shape': (10, 5), 'dtype': 'float32'},
            ...              'biases': {'shape': (10,), 'dtype': 'float32'},
            ...              'label': {'shape': None, 'dtype': 'str'},
            ...              'vector': {'dtype': 'float32'}}
            >>> n_walkers = 3
            >>> state = State(param_dict=attr_dict, n_walkers=n_walkers)
            >>> tensor_dict = state.params_to_arrays(attr_dict, n_walkers)
            >>> tensor_dict.keys()
            dict_keys(['weights', 'biases', 'label', 'vector'])
            >>> tensor_dict['weights'].shape
            (3, 10, 5)
            >>> tensor_dict['biases'].shape
            (3, 10)
            >>> tensor_dict['label']
            [None, None, None]
            >>> tensor_dict['vector'].shape
            (3,)
        """
        tensor_dict = {}
        for key, val in param_dict.items():
            val = copy.deepcopy(val)
            shape = val.pop("shape", -1)  # If shape is not specified, assume it's a scalar vector.
            if shape is None or key in self.list_names:
                # If shape is None, assume it's not a tensor but a list.
                value = [None] * n_walkers
            else:  # Initialize the tensor to zeros. Assumes dtype is a valid argument.
                if shape == -1:
                    shape = (n_walkers,)
                else:
                    shape = (
                        (n_walkers, shape) if isinstance(shape, int) else ((n_walkers,) + shape)
                    )
                try:
                    value = API.zeros(shape, **val)
                except TypeError as e:
                    print("FAIL", key)
                    raise TypeError(f"Invalid dtype for {key}: {val['dtype']}") from e

            tensor_dict[key] = value
        return tensor_dict

    def _parse_value(self, name: str, value: Any) -> Any:
        """
        Ensure that the input value has correct dimensions and shape for a given attribute.

        Args:
            name (str): Name of the attribute.
            value (Any): New value to set to the attribute.

        Returns:
            Any: Parsed and validated value of the new state element.

        Example:
            >>> s = State(2, {'name': {'shape': (3, 4), "dtype": int}})
            >>> parsed_val = s._parse_value('name', [[1, 2, 3, 4], [5, 6, 7, 8], [9, 10, 11, 12]])
            >>> parsed_val
            array([[ 1,  2,  3,  4],
                   [ 5,  6,  7,  8],
                   [ 9, 10, 11, 12]])
        """
        if name in self.list_names:
            assert isinstance(value, list), (name, value)
            return value
        dtype = type(value[0])
        if name in self.names:
            dtype = self._param_dict[name].get("dtype", dtype)
        else:
            raise NotImplementedError(f"Name {name} must be present in self.names")
        try:
            tensor_val = tensor(value, dtype=dtype)
        except ValueError as e:
            raise ValueError(f"Name {name} failed conversion to tensor:") from e
        except TypeError as e:
            raise TypeError(
                (f"Name {name} with type {dtype} and value {value} failed conversion to tensor:"),
            ) from e
        return tensor_val


class SwarmState(State):
    """
    A dictionary-style container for storing the current state of the swarm. It allows you
    to update the status and metadata of the walkers in the swarm.

    The keys of the instance must be its attributes. The attribute value can be of \
    any type (tensor, list or any python object). Lists and Tensors can have different \
    len than ``n_walkers`` if necessary, but tensors should have the same number of \
    rows as walkers (whether active or not).

    Args:
        n_walkers (int): Number of walkers
        param_dict (StateDict): Dictionary defining the attributes of the tensors.

    Example:
        >>> param_dict = {"x": {"shape": (3, 4), "dtype": int}}
        >>> s = SwarmState(n_walkers=100, param_dict=param_dict)
    """

    def __init__(self, n_walkers: int, param_dict: StateDict):
        """
        Initialize a :class:`SwarmState`.

        Args:
             n_walkers: The number of items in the first dimension of the tensors.
             param_dict: Dictionary defining the attributes of the tensors.

        """
        self._clone_names = set(k for k, v in param_dict.items() if v.get("clone"))
        super(SwarmState, self).__init__(n_walkers=n_walkers, param_dict=param_dict)
        self._actives = judo.ones(self.n_walkers, dtype=judo.dtype.bool)
        self._n_actives = int(self.n_walkers)

    @property
    def clone_names(self) -> Set[str]:
        """Return the name of the attributes that will be copied when a walker clones."""
        return self._clone_names

    @property
    def actives(self) -> Tensor:
        """Get the active walkers indices."""
        return self._actives

    @property
    def n_actives(self) -> int:
        """Get the number of active walkers."""
        return self._n_actives

    def clone(self, will_clone, compas_clone, clone_names):
        """Clone all the stored data according to the provided index."""
        for name in clone_names:
            values = self[name]
            if name in self.tensor_names:
                self[name][will_clone] = values[compas_clone][will_clone]
            else:
                self[name] = [
                    values[comp] if wc else val
                    for val, comp, wc in zip(values, compas_clone, will_clone)
                ]

    def export_walker(self, index: int, names: Optional[List[str]] = None):
        """
        Export the data of a walker at index `index` as a dictionary.

        Args:
            index (int): The index of the target walker.
            names (Optional[List[str]], optional): The list of attribute names
                to be included in the output. If None,
                all attributes will be included. Defaults to None.

        Returns:
            Dict[str, Union[Tensor, numpy.ndarray]]: A dictionary containing the \
            requested attributes and their corresponding values for the specified walker.

        Examples:
            >>> param_dict = {"x": {"shape": (3, 4), "dtype": int}}
            >>> s = SwarmState(n_walkers=10, param_dict=param_dict)
            >>> s.reset()
            >>> walker_dict = s.export_walker(0, names=["x"])
            >>> print(walker_dict)  # doctest: +NORMALIZE_WHITESPACE
            {'x': array([[[0, 0, 0, 0],
                          [0, 0, 0, 0],
                          [0, 0, 0, 0]]])}
        """
        names = names if names is not None else self.keys()
        return {
            k: (v[[index]] if k in self.tensor_names else [v[index]])
            for k, v in self.items()
            if k in names
        }

    def import_walker(self, data: Dict[str, Tensor], index: int = 0) -> None:
        """Takes data dictionary and imports it into state at indice `index`.

        Args:
            data (Dict): Dictionary containing the data to be imported.
            index (int, optional): Walker index to receive the data. Defaults to 0.

        Examples:
            >>> param_dict = {"x": {"shape": (3, 4), "dtype": int}}
            >>> s = SwarmState(n_walkers=10, param_dict=param_dict)
            >>> s.reset()
            >>> data = {"x": judo.ones((3, 4), dtype=int)}
            >>> s.import_walker(data, index=0)
            >>> s.get("x")[0, 0, :3]
            array([1, 1, 1])
        """
        for name, tensor_ in data.items():
            if name in self.tensor_names:
                self[name][[index]] = judo.copy(tensor_)
            else:
                self[name][index] = copy.deepcopy(tensor_[0])

    def reset(self, root_walker: Optional[Dict[str, Tensor]] = None) -> None:
        """
        Completely resets both current and history data that have been held in state.

        Optionally can take a root value to reset individual attributes.

        Args:
            root_walker (Optional[Dict[str, Tensor]], optional): The initial state when resetting.

        Examples:
            >>> param_dict = {"x": {"shape": (3, 4), "dtype": int}}
            >>> s = SwarmState(n_walkers=10, param_dict=param_dict)
            >>> s.reset()
            >>> walker_dict = s.export_walker(0, names=["x"])
            >>> print(walker_dict["x"].shape)  # doctest: +NORMALIZE_WHITESPACE
            (1, 3, 4)
        """
        self._actives = judo.ones(self.n_walkers, dtype=judo.dtype.bool)
        self._n_actives = int(self.n_walkers)
        super(SwarmState, self).reset()
        if root_walker:
            for name, array in root_walker.items():
                if name in self.tensor_names:
                    self[name][:] = judo.copy(array)
                else:
                    self[name] = copy.deepcopy([array[0] for _ in range(self.n_walkers)])

    def get(self, name: str, default=None, raise_error: bool = True, inactives: bool = False):
        """
        Get an attribute by key and return the default value if it does not exist.

        Args:
            name: Attribute to be recovered.
            default: Value returned in case the attribute is not part of state.
            raise_error: If True, raise AttributeError if name is not present in states.
            inactives: Whether to update the walkers marked as inactive.

        Returns:
            Target attribute if found in the instance, otherwise returns the
            default value.

        Examples:
            >>> param_dict = {"x": {"shape": (3, 4), "dtype": int}}
            >>> s = SwarmState(n_walkers=10, param_dict=param_dict)
            >>> s.reset()
            >>> print(s.get("x").shape)
            (10, 3, 4)


        """
        value = super(SwarmState, self).get(name=name, default=default, raise_error=raise_error)
        if inactives or name not in self.names:
            return value
        try:
            return (
                value[self.actives]
                if name not in self.list_names
                else [x for x, active in zip(value, self.actives) if active]
            )
        except Exception as e:
            print("NAME", name, self.actives.shape, value.shape)
            raise e

    def _update_active_list(self, current_vals, new_vals):
        """Update the list of active walkers."""
        new_ix = 0
        updated_vals = []
        for active, curr in zip(self.actives, current_vals):
            if active:
                updated_vals.append(new_vals[new_ix])
                new_ix += 1
            else:
                updated_vals.append(curr)
        return updated_vals

    def update(
        self,
        other: Union["SwarmState", Dict[str, Tensor]] = None,
        inactives: bool = False,
        **kwargs,
    ) -> None:
        """
        Modify the data stored in the SwarmState instance.

        Existing attributes will be updated, and new attributes will be created if needed.

        Args:
            other: State class that will be copied upon update.
            inactives: Whether to update the walkers marked as inactive.
            **kwargs: It is possible to specify the update as name value attributes,
                where name is the name of the attribute to be updated, and value
                is the new value for the attribute.
        Returns:
            None.

        Examples:
            >>> param_dict = {"x": {"shape": (3, 4), "dtype": int}}
            >>> s = SwarmState(n_walkers=10, param_dict=param_dict)
            >>> s.reset()
            >>> s.update(x=judo.ones((10, 3, 4), dtype=int))
            >>> print(s.get("x")[0,0,0])
            1
        """
        if other is not None:
            return super(SwarmState, self).update(other=other, **kwargs)
        # FIXME (guillemdb): This does not make sense because it executes only if other is None.
        new_values = other.to_dict() if isinstance(other, SwarmState) else (other or {})
        new_values.update(kwargs)
        for name, val in new_values.items():
            if name == "actives":
                continue
            val = self._parse_value(name, val)
            is_actives_vector = (len(val) == self._n_actives) or (
                judo.is_tensor(val) and val.shape[0] == self._n_actives
            )
            if is_actives_vector and hasattr(self, name) and not inactives:
                all_vals = getattr(self, name)
                if isinstance(all_vals, list):
                    new_vals = self._update_active_list(all_vals, val)
                    assert len(new_vals) == self.n_walkers
                    setattr(self, name, new_vals)
                else:
                    try:
                        all_vals[self.actives] = val
                    except RuntimeError:
                        all_vals[self.actives] = judo.copy(val)
                    setattr(self, name, all_vals)

            elif inactives or self.n_actives == self.n_walkers or len(val) == self.n_walkers:
                setattr(self, name, val)
            else:
                raise ValueError(f"failed to setup data {name} {val.shape}")

    def update_actives(self, actives) -> None:
        """
        Set the walkers marked as active.

        Example:
            To set the first 10 walkers as active, call the function with a tensor of size
            equal to the number of walkers, where the first ten elements are `True`
            and the remaining elements are `False`:

            >>> param_dict = {"vector":{"dtype":int}}
            >>> s = SwarmState(n_walkers=20, param_dict=param_dict)
            >>> active_walkers = np.concatenate([np.ones(10), np.zeros(10)]).astype(bool)
            >>> s.update_actives(active_walkers)

            This will mark those walkers as active, and any attribute updated with inactives=False
            (this is the default) will only modify the data from those walkers.
        """
        self._actives = actives
        self._n_actives = int(actives.sum())
